# Calculates different variables from plot3d data
# Equations can be found here: https://www.grc.nasa.gov/WWW/winddocs/towne/plotc/plotc_p3d.html
# TODO: Add more variables
#  Enthalpies, Vorticity, Entropy, Turbulence Parameters, Gradients, Move metrics here?
#  Total quantities
import numpy as np


class Variables:
    """
    Module to compute flow variables

    ...

    Attributes
    ----------
    Input:
        flow : src.io.plot3dio.FlowIO or src.streamlines.interpolation.Interpolation or similar
            object with q --> flow data
        gamma : float
            default is 1.4; can specify
    Output:
        velocity : ndarray
            velocity of the flow at nodes; shape of the flow.q (ni x nj x nk x 3 x nb)
        velocity_magnitude : ndarray
            magnitude of velocity at nodes; shape is (ni x nj x nk x nb)
        temperature : ndarray
            temperature at nodes; shape is (ni x nj x nk x nb)
        pressure : ndarray
            pressure at nodes; shape is (ni x nj x nk x nb)

    Methods
    -------
    compute_velocity()
        computes the velocity and velocity_magnitude
    compute_temperature()
        computes the temperature
    compute_pressure()
        computes the pressure
    compute()
        computes all the variables available

    ...

    Example:
    -------
        variables = Variables(flow)  # Assume flow object is pre-defined
        variables.compute_velocity()  # returns velocity attribute
        variables.compute_temperature()  # fills up the temperature attribute
        variables.compute()  # computes all the attributes available

    """

    def __init__(self, flow, gamma=1.4, gas_constant=287.052874):
        self.flow = flow
        self.gamma = gamma
        self.gas_constant = gas_constant
        self.density = flow.q[..., 0, :]  # q0
        self.velocity = None
        self.mach = None
        self.velocity_magnitude = None
        self.temperature = None
        self.pressure = None
        self.viscosity = None

    def compute_velocity(self):
        """
        Function to compute velocity and velocity magnitude
        :return: None
        """
        # velocity = [q1, q2, q3] / q0
        self.velocity = self.flow.q[..., 1:4, :] / (self.flow.q[..., 0, :, None])
        self.velocity_magnitude = (self.velocity[..., 0, :]**2 + self.velocity[..., 1, :]**2 + self.velocity[..., 2, :]**2)**0.5

        return

    def compute_temperature(self):
        """
        Function to compute temperature.
        This computes velocity first
        :return: None
        """
        self.compute_velocity()
        _q4 = self.flow.q[..., 4, :]
        self.temperature = (self.gamma - 1) * (_q4/self.density - self.velocity_magnitude**2/2) / self.gas_constant

        return

    def compute_mach(self):
        """
        Function to compute local mach number
        Returns: None
        """
        self.compute_temperature()
        self.mach = self.velocity_magnitude / np.sqrt(self.gamma * self.gas_constant * self.temperature)

        return

    def compute_pressure(self):
        """
        Function to compute pressure.
        This computes velocity and temperature first
        :return: None
        """
        self.compute_mach()
        self.pressure = self.density * self.temperature * self.gas_constant

        return

    def compute_viscosity(self, law='keyes'):
        """
        Function to compute viscosity

        Viscosity of air as a function of temperature
        Args:
            _temperature: in Kelvin provided by default in the integration class
            law: 'sutherland' or 'keyes' -- defaults to 'keyes'
        :return: None

        """
        match law:
            case 'sutherland':
                # Sutherland's viscosity law
                # All temperatures must be in kelvin
                # Formula from cfd-online
                _c1 = 1.716e-5 * (273.15 + 110.4) / 273.15**1.5
                self.viscosity = _c1 * self.temperature**1.5 * 0.4 / (self.temperature + 110.4)
            case 'keyes':
                # New formula from keyes et al.
                a0, a, a1 = 1.488, 122.1, 5.0
                _tau = 1/self.temperature
                self.viscosity = a0 * self.temperature**0.5 * 10**-6 / (1 + a * _tau / 10 ** (a1 * _tau))
            case _:
                raise ValueError('Viscosity law not supported')

        return

    def compute_drag_coefficient(self, _re=None, _mach=None, _model='stokes'):
        """
        Coefficient of drag for a spherical particle

        Args:
            _re : Relative Reynolds Number
            _mach : Relative Mach Number
            _model : Drag Model Name
                Available models are 'sphere', 'stokes', 'oseen', 'schiller_nauman',
                'cunningham'

        Returns:
            coefficient of drag based on local flow/particle properties

        """
        match _model:
            case 'zero-drag':
                # zero drag model to simulate fluid
                return 0

            case 'sphere':
                # ref: Fluid Mechanics, Frank M. White
                # This was decided by trail-and-error from VISUAL3 code
                if _re <= 1e-9:
                    return 0
                if _re < 1e-3:
                    return 24 / _re
                if 1e-3 <= _re < 0.45:
                    return 24 / _re * (1 + 3 * _re / 16)
                if 0.45 <= _re < 1:
                    # Same as above due to lack of data
                    return 24 / _re * (1 + 3 * _re / 16)
                if 1 <= _re < 800:
                    return 24 / _re * (1 + _re ** (2 / 3) / 6)
                if 800 <= _re < 3e5:
                    return 0.44
                if 3e5 <= _re < 4e5:
                    # Same as above due to lack of data
                    return 0.44
                if _re >= 4e5:
                    return 0.07

            case 'stokes':
                # Stokes Drag; for creeping flow regime; Re << 1
                if _re <= 1e-9:
                    return 0
                else:
                    return 24/_re

            case 'melling':
                # The popular melling correction
                if _re <= 1e-9:
                    return 0
                else:
                    knd = _mach / _re * np.sqrt(np.pi*self.gamma/2)
                    return 24/_re * (1 + knd)**-1

            case 'oseen':
                # Oseen's model; for creeping flow regime; Re < 1
                if _re <= 1e-9:
                    return 0
                else:
                    return 24/_re * (1 + 3/16 * _re)

            case 'schiller-nauman':
                # Schiller and Nauman's model; for Re <~ 200 & M <~ 0.25
                if _re <= 1e-9:
                    return 0
                else:
                    return 24/_re * (1 + 0.15 * _re**0.687)

            case 'cunningham':
                # Cunningham model; for Re << 1; M << 1; Kn <~ 0.1
                # Knudsen number
                if _re <= 1e-9:
                    return 0
                if _re <= 1:
                    _kn = _mach / _re * np.sqrt(self.gamma * np.pi/2)
                    return 24/_re * (1 + 4.5*_kn)**-1
                if _re > 1:
                    _kn = _mach / np.sqrt(_re)
                    return 24/_re * (1 + 4.5*_kn)**-1

            case 'henderson':
                # Henderson model; for all flow regimes
                # Simplified by ignoring sphere temperature
                if _re < 1e-9:
                    return 0

                # For Mach < 1
                _s = _mach * np.sqrt(self.gamma/2)
                _f1 = 24 * (_re + _s * (5.89688 * np.exp(-0.247 * _re/_s)))**-1
                _f2 = np.exp(-0.5*_mach/np.sqrt(_re)) * \
                      ((4.5 + 0.38*(0.03*_re + 0.48*np.sqrt(_re))) / (1 + 0.03*_re + 0.48*np.sqrt(_re)) +
                       0.1*_mach**2 + 0.2*_mach**8)
                _f3 = (1 - np.exp(-_mach/_re))*0.6*_s
                _cd1 = _f1 + _f2 + _f3
                if _mach < 1:
                    return _cd1

                # For Mach >= 1.75
                _mach_inf = _mach
                _re_inf = _re
                _s_inf = _mach_inf * np.sqrt(self.gamma/2)
                _g1 = 0.9 + 0.34/_mach_inf**2
                _g2 = 1.86 * np.sqrt(_mach_inf/_re_inf) * (2 + 2/_s_inf**2 + 1.058/_s_inf - 1/_s_inf**4)
                _g3 = 1 + 1.86 * np.sqrt(_mach_inf/_re_inf)
                _cd2 = (_g1 + _g2) / _g3
                if _mach >= 1.75:
                    return _cd2

                # For 1 <= Mach < 1.75; linear interpolation
                if 1 <= _mach < 1.75:
                    return _cd1 + 4/3 * (_mach_inf - 1) * (_cd2 - _cd1)

            case 'subramaniam-balachandar':
                # Model from their new book
                if _re < 1e-9:
                    return 0

                if _re < 0.5:
                    # Stokes
                    return 24/_re

                if _re < 20:
                    # Clift
                    return 24/_re * (1 + 0.1315 * _re**(0.82-0.05*np.log10(_re)))

                if _re < 800:
                    # Schiller-Naumann
                    return 24/_re * (1 + 0.15 * _re**0.687)

                if _re < 3e5:
                    # Clift-Gauvin
                    return 24/_re * (1 + 0.15 * _re**0.687 + 0.42/24 * _re * (1 + 4.25e4 * _re**(-1.16))**-1)

            case 'loth':
                # Loth's model; for all flow regimes
                if _re < 1e-9:
                    return 0

                if _re < 45:
                    # Rarefraction dominated domain
                    import math
                    _s = _mach * np.sqrt(self.gamma/2)
                    _cd_fm = (1 + 2 * _s**2) * np.exp(-_s**2) / (_s**3 * np.pi**0.5) + \
                             (4*_s**4 + 4*_s**2 - 1) * math.erf(_s) / (2*_s**4) + 2 * np.pi**0.5 / (3 * _s)
                    _cd_fm_re = _cd_fm / (1 + (_cd_fm/1.63 - 1) * (_re/45)**0.5)
                    _kn = (np.pi * self.gamma / 2)**0.5 * _mach / _re
                    _f_kn = (1 + _kn * (2.514 + 0.8 * np.exp(-0.55/_kn)))**-1
                    _cd_kn_re = 24/_re * (1 + 0.15 * _re**0.687) * _f_kn
                    _cd = (_cd_kn_re + _mach**4 * _cd_fm_re) / (1 + _mach**4)
                    return _cd

                if _re == 45:
                    return 1.63

                if _re > 45:
                    # compression-dominated regime
                    if _mach <= 1.45:
                        _cm = 5/3 + 2/3 * np.tanh(3 * np.log(_mach - 0.1))
                    if _mach > 1.45:
                        _cm = 2.044 + 0.2 * np.exp(-1.8 * (np.log(_mach/1.5))**2)
                    if _mach <= 0.89:
                        _gm = 1 - 1.525 * _mach**4
                    if _mach > 0.89:
                        _gm = 2e-4 + 8e-4 * np.tanh(12.77 * (_mach - 2.02))
                    _hm = 1 - 0.258 * _cm / (1 + 514 * _gm)
                    _cd = 24/_re * (1 + 0.15 * _re**0.687) * _hm + 0.42 * _cm / (1 + 42000 * _gm / _re**1.16)
                    return _cd

                return

    def compute(self):
        # implicitly runs compute_velocity() and compute_temperature()
        """
        Function to compute all the attributes in the class
        :return: None
        """
        self.compute_pressure()
        # This computes viscosity using the keyes formula
        # To compute using sutherland's law, call compute_viscosity() separately, which will update the attribute
        self.compute_viscosity()



import unittest
import matplotlib.pyplot as plt


class TestCreateParticles(unittest.TestCase):
    def test_create_particles(self):
        from src.sypivlib.function.dataio import GridIO, FlowIO
        from src.sypivlib.sypiv.create_particles import Particle, LaserSheet, CreateParticles

        # Read-in the grid and flow file
        grid = GridIO('../data/shocks/shock_test.sb.sp.x')
        grid.read_grid()
        grid.compute_metrics()
        flow = FlowIO('../data/shocks/shock_test.sb.sp.q')
        flow.read_flow()

        # Set particle data
        p = Particle()
        p.min_dia = 144e-9
        p.max_dia = 573e-9
        p.mean_dia = 281e-9
        p.std_dia = 97e-9
        p.density = 810
        p.n_concentration = 500
        p.compute_distribution()
        # print(p.particle_field)

        # Read-in the laser sheet
        laser = LaserSheet(grid)
        laser.position = 0.0009
        laser.thickness = 0.0001  # Adjusted based on grid thickness
        laser.pulse_time = 1e-7
        laser.compute_bounds()
        # print(laser.width)

        # Create particle locations array
        ia_bounds = [None, None, None, None]
        loc = CreateParticles(grid, flow, p, laser, ia_bounds)
        loc.ia_bounds = [0, 0.003, 0, 0.001]
        loc.in_plane = 90
        loc.compute_locations()
        loc.compute_locations2_serial()

        # Sample code to plot particle locations and relative diameters
        _in_plane = int(p.n_concentration * loc.in_plane * 0.01)
        # plot in-plane particle locations
        plt.scatter(loc.locations[:_in_plane, 0], loc.locations[:_in_plane, 1],
                    s=10*loc.locations[:_in_plane, 3]/p.min_dia, c='g')
        # plot out-of-plane locations
        plt.scatter(loc.locations[_in_plane:, 0], loc.locations[_in_plane:, 1],
                    s=10*loc.locations[_in_plane:, 3]/p.min_dia, c='r')
        plt.xlim([-0.0001, 0.004])
        plt.ylim([0, 0.0019])
        plt.title('Particles created on the given IA for the first snap')

        # plt.show()

        # plot in-plane particle locations
        plt.figure()
        plt.scatter(loc.locations2[:_in_plane, 0], loc.locations2[:_in_plane, 1],
                    s=10 * loc.locations2[:_in_plane, 3] / p.min_dia, c='g')
        # plot out-of-plane locations
        plt.scatter(loc.locations2[_in_plane:, 0], loc.locations2[_in_plane:, 1],
                    s=10 * loc.locations2[_in_plane:, 3] / p.min_dia, c='r')
        plt.xlim([-0.0001, 0.004])
        plt.ylim([0, 0.0019])
        plt.title('Particles created on the given IA for the second snap')

        plt.show()


if __name__ == '__main__':
    unittest.main()



# Integrate from given point to produce streamlines
import numpy as np
from .interpolation import Interpolation
from .search import Search
from .variables import Variables
from scipy.optimize import fsolve


class Integration:

    def __init__(self, interp):
        self.interp = interp
        self.ppoint = None
        self.cpoint = None
        self.rk4_bool = False

    def __str__(self):
        doc = "This instance uses data from " + self.interp.flow.filename + \
              " and integrates based on the given time step"
        return doc

    def compute(self, method='p-space', time_step=1e-6):
        match method:
            case 'p-space':
                # For p-space algos; the point-in-domain check was done in search
                if self.interp.idx.ppoint is None:
                    self.ppoint = None
                    return self.ppoint

                # Compute required variables from plot3d data
                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                # Integration for one time step
                self.ppoint = self.interp.idx.ppoint + q_interp.velocity.reshape(3) * time_step

                return self.ppoint

            case 'c-space':
                # Get inverse Jacobian from the interpolation class
                # Using cell node data. For more accurate calculation refer to cRK4 method
                _J_inv = self.interp.idx.grid.m2[self.interp.idx.cell[0, 0], self.interp.idx.cell[0, 1],
                                                 self.interp.idx.cell[0, 2], :, :, self.interp.idx.block]

                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                p_velocity = q_interp.velocity.reshape(3)
                c_velocity = np.matmul(_J_inv, p_velocity)
                self.cpoint = self.interp.idx.cpoint + c_velocity * time_step

                # For c-space the point in-domain check is done after integration
                if not np.all([0, 0, 0] <= self.cpoint) or not np.all(
                        self.cpoint + 1 < [self.interp.idx.grid.ni[self.interp.idx.block],
                                           self.interp.idx.grid.nj[self.interp.idx.block],
                                           self.interp.idx.grid.nk[self.interp.idx.block]]):
                    self.cpoint = None
                    return self.cpoint

                return self.cpoint

            case 'pRK2':
                """
                This is a straight forward RK2 integration. Search for the point,
                Interpolate the data to the point, Compute required variables,
                Perform RK4 integration!
                """

                def _rk2_step(self, x):
                    """

                    Args:
                        self:
                        x: ndarray
                            point in p-space

                    Returns:
                        k: ndarray
                            interim RK4 variables

                    """
                    idx = Search(self.interp.idx.grid, x)
                    idx.compute(method='p-space')
                    # For p-space algos; the point-in-domain check was done in search
                    if idx.ppoint is None: return None, None
                    interp = Interpolation(self.interp.flow, idx)
                    interp.adaptive = self.interp.adaptive
                    interp.rbf_kernel = self.interp.rbf_kernel
                    interp.compute(method=self.interp.method)
                    q_interp = Variables(interp)
                    q_interp.compute_velocity()
                    u = q_interp.velocity.reshape(3)
                    k = time_step * u
                    return u, k

                # Start RK4 for p-space
                # For p-space algos; the point-in-domain check was done in search
                x0 = self.interp.idx.ppoint
                if x0 is None: return None, None
                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                u0 = q_interp.velocity.reshape(3)
                k0 = time_step * u0
                x1 = x0 + k0

                u1, k1 = _rk2_step(self, x1)
                if k1 is None: return None, None

                x_new = x0 + 1/2 * (k0 + k1)
                u_new, k_new = _rk2_step(self, x_new)
                if k_new is None: return None, None

                self.ppoint = x_new

                return self.ppoint, u_new

            case 'cRK2':
                '''
                This is a block-wise integration. Everytime the point gets out
                a pRK4 is run to find the new block the point is located in.
                That particular step is done in streamlines algorithm.

                All the points, x0, x1... are in c-space

                Point location is known in c-space, avoiding search.
                Interpolates data to the point
                RK4 integration is performed!
                '''

                def _rk2_step(self, x):
                    """

                    Args:
                        self:
                        x: ndarray
                            point in c-space

                    Returns:
                        k: ndarray
                            interim RK2 variables

                    """
                    idx = Search(self.interp.idx.grid, x)
                    idx.block = self.interp.idx.block
                    idx.c2p(x)  # This will change the cell attribute
                    # In-domain check is done in search
                    if idx.cpoint is None:
                        self.cpoint = None
                        self.ppoint = None
                        return None, None, None
                    interp = Interpolation(self.interp.flow, idx)
                    interp.adaptive = self.interp.adaptive
                    interp.rbf_kernel = self.interp.rbf_kernel
                    interp.compute(method=self.interp.method)
                    _J_inv = interp.J_inv
                    q_interp = Variables(interp)
                    q_interp.compute_velocity()
                    p_velocity = q_interp.velocity.reshape(3)
                    c_velocity = np.matmul(_J_inv, p_velocity)
                    k = time_step * c_velocity
                    return k, p_velocity, c_velocity

                x0 = self.interp.idx.cpoint
                _J_inv = self.interp.J_inv
                if x0 is None:
                    return None, None, None

                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                p_velocity = q_interp.velocity.reshape(3)
                c_velocity = np.matmul(_J_inv, p_velocity)
                k0 = time_step * c_velocity
                x1 = x0 + k0

                k1, pv1, cv1 = _rk2_step(self, x1)
                if k1 is None:
                    return None, None, None

                x_new = x0 + 1/2 * (k0 + k1)
                _, pv_new, cv_new = _rk2_step(self, x_new)

                self.cpoint = x_new

                return self.cpoint, pv_new, cv_new

            case 'pRK4':
                """
                This is a straight forward RK4 integration. Search for the point,
                Interpolate the data to the point, Compute required variables,
                Perform RK4 integration!
                """

                def _rk4_step(self, x):
                    """

                    Args:
                        self:
                        x: ndarray
                            point in p-space

                    Returns:
                        k: ndarray
                            interim RK4 variables

                    """
                    idx = Search(self.interp.idx.grid, x)
                    idx.compute(method='p-space')
                    # For p-space algos; the point-in-domain check was done in search
                    if idx.ppoint is None:
                        return None, None
                    interp = Interpolation(self.interp.flow, idx)
                    interp.adaptive = self.interp.adaptive
                    interp.rbf_kernel = self.interp.rbf_kernel
                    interp.compute(method=self.interp.method)
                    q_interp = Variables(interp)
                    q_interp.compute_velocity()
                    u = q_interp.velocity.reshape(3)
                    k = time_step * u
                    return u, k

                # Start RK4 for p-space
                # For p-space algos; the point-in-domain check was done in search
                x0 = self.interp.idx.ppoint
                if x0 is None:
                    return None, None
                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                u0 = q_interp.velocity.reshape(3)
                k0 = time_step * u0
                x1 = x0 + 0.5 * k0

                u1, k1 = _rk4_step(self, x1)
                if k1 is None:
                    return None, None
                x2 = x0 + 0.5 * k1

                u2, k2 = _rk4_step(self, x2)
                if k2 is None:
                    return None, None
                x3 = x0 + k2

                u3, k3 = _rk4_step(self, x3)
                if k3 is None:
                    return None, None
                x_new = x0 + 1/6 * (k0 + 2*k1 + 2*k2 + k3)
                u_new, _ = _rk4_step(self, x_new)

                self.ppoint = x_new

                return self.ppoint, u_new

            case 'cRK4':
                '''
                This is a block-wise integration. Everytime the point gets out
                a pRK4 is run to find the new block the point is located in.
                That particular step is done in streamlines algorithm.
                
                All the points, x0, x1... are in c-space
                
                Point location is known in c-space, avoiding search.
                Interpolates data to the point
                RK4 integration is performed!
                '''

                def _rk4_step(self, x):
                    """

                    Args:
                        self:
                        x: ndarray
                            point in c-space

                    Returns:
                        k: ndarray
                            interim RK4 variables

                    """
                    idx = Search(self.interp.idx.grid, x)
                    idx.block = self.interp.idx.block
                    idx.c2p(x)  # This will change the cell attribute
                    # In-domain check is done in search
                    if idx.cpoint is None:
                        self.cpoint = None
                        self.ppoint = None
                        return None, None, None
                    interp = Interpolation(self.interp.flow, idx)
                    interp.adaptive = self.interp.adaptive
                    interp.rbf_kernel = self.interp.rbf_kernel
                    interp.compute(method=self.interp.method)
                    _J_inv = interp.J_inv
                    q_interp = Variables(interp)
                    q_interp.compute_velocity()
                    p_velocity = q_interp.velocity.reshape(3)
                    c_velocity = np.matmul(_J_inv, p_velocity)
                    k = time_step * c_velocity
                    return k, p_velocity, c_velocity

                x0 = self.interp.idx.cpoint
                _J_inv = self.interp.J_inv
                if x0 is None:
                    return None, None, None

                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                p_velocity = q_interp.velocity.reshape(3)
                c_velocity = np.matmul(_J_inv, p_velocity)
                k0 = time_step * c_velocity
                x1 = x0 + 0.5 * k0

                k1, pv1, cv1 = _rk4_step(self, x1)
                if k1 is None:
                    return None, None, None
                x2 = x0 + 0.5 * k1

                k2, pv2, cv2 = _rk4_step(self, x2)
                if k2 is None:
                    return None, None, None
                x3 = x0 + k2

                k3, pv3, cv3 = _rk4_step(self, x3)
                if k3 is None:
                    return None, None, None
                x_new = x0 + 1/6 * (k0 + 2*k1 + 2*k2 + k3)
                _, pv_new, cv_new = _rk4_step(self, x_new)

                self.cpoint = x_new

                return self.cpoint, pv_new, cv_new

            case 'unsteady-pRK4':
                """
                This is a straight forward RK4 integration. Search for the point,
                Interpolate the data to the point, Compute required variables,
                Perform RK4 integration!
                """

                def _rk4_step(self, x):
                    """

                    Args:
                        self:
                        x: ndarray
                            point in p-space

                    Returns:
                        k: ndarray
                            interim RK4 variables

                    """
                    idx = Search(self.interp.idx.grid, x)
                    idx.compute(method='p-space')
                    # For p-space algos; the point-in-domain check was done in search
                    if idx.ppoint is None: return None, None
                    interp = Interpolation(self.interp.flow, idx)
                    interp.adaptive = self.interp.adaptive
                    interp.rbf_kernel = self.interp.rbf_kernel
                    interp.time = self.interp.time
                    interp.flow_old = self.interp.flow_old
                    interp.compute(method='unsteady-rbf-p-space')
                    q_interp = Variables(interp)
                    q_interp.compute_velocity()
                    u = q_interp.velocity.reshape(3)
                    k = time_step * u
                    return u, k

                # Start RK4 for p-space
                # For p-space algos; the point-in-domain check was done in search
                x0 = self.interp.idx.ppoint
                if x0 is None: return None, None
                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                u0 = q_interp.velocity.reshape(3)
                k0 = time_step * u0
                x1 = x0 + 0.5 * k0

                u1, k1 = _rk4_step(self, x1)
                if k1 is None: return None, None
                x2 = x0 + 0.5 * k1

                u2, k2 = _rk4_step(self, x2)
                if k2 is None: return None, None
                x3 = x0 + k2

                u3, k3 = _rk4_step(self, x3)
                if k3 is None: return None, None
                x_new = x0 + 1 / 6 * (k0 + 2 * k1 + 2 * k2 + k3)
                u_new, _ = _rk4_step(self, x_new)

                self.ppoint = x_new

                return self.ppoint, u_new

    def compute_ppath(self, diameter=1e-6, density=1000, velocity=None,
                      method='pRK4', time_step=1e-4, drag_model='stokes'):

        def _drag_constant(_re, _q_interp=None, _mach=None, _mu=None, _model=drag_model):
            """
            Coefficient of drag for a spherical particle

            Args:
                _re : Relative Reynolds Number
                _mach : Relative Mach Number
                _model : Drag Model Name
                    Available models are 'sphere', 'stokes', 'oseen', 'schiller_nauman',
                    'cunningham'

            Returns:
                coefficient of drag based on local flow/particle properties

            """
            _gamma = q_interp.gamma
            match _model:
                case 'zero-drag':
                    # zero drag model to simulate fluid
                    return 0

                case 'sphere':
                    # ref: Fluid Mechanics, Frank M. White
                    # This was decided by trail-and-error from VISUAL3 code
                    if _re <= 1e-9:
                        return 0
                    if _re < 1e-3:
                        return 24 / _re
                    if 1e-3 <= _re < 0.45:
                        return 24 / _re * (1 + 3 * _re / 16)
                    if 0.45 <= _re < 1:
                        # Same as above due to lack of data
                        return 24 / _re * (1 + 3 * _re / 16)
                    if 1 <= _re < 800:
                        return 24 / _re * (1 + _re ** (2 / 3) / 6)
                    if 800 <= _re < 3e5:
                        return 0.44
                    if 3e5 <= _re < 4e5:
                        # Same as above due to lack of data
                        return 0.44
                    if _re >= 4e5:
                        return 0.07

                case 'stokes':
                    # Stokes Drag; for creeping flow regime; Re << 1
                    if _re <= 1e-9:
                        return 0
                    else:
                        return 24/_re

                case 'melling':
                    # The popular melling correction
                    if _re <= 1e-9:
                        return 0
                    else:
                        knd = _mach / _re * np.sqrt(np.pi*_gamma/2)
                        return 24/_re * (1 + knd)**-1

                case 'melling-2':
                    # The popular melling correction
                    if _re <= 1e-9:
                        return 0
                    else:
                        knd = _mach / _re * np.sqrt(np.pi*_gamma/2)
                        return 24/_re * (1 + 2.7*knd)**-1

                case 'oseen':
                    # Oseen's model; for creeping flow regime; Re < 1
                    if _re <= 1e-9:
                        return 0
                    else:
                        return 24/_re * (1 + 3/16 * _re)

                case 'schiller-nauman':
                    # Schiller and Nauman's model; for Re <~ 200 & M <~ 0.25
                    if _re <= 1e-9:
                        return 0
                    else:
                        return 24/_re * (1 + 0.15 * _re**0.687)

                case 'cunningham':
                    # Cunningham model; for Re << 1; M << 1; Kn <~ 0.1
                    # Knudsen number
                    # _r = _q_interp.density * _q_interp.velocity_magnitude * diameter / _mu
                    if _re <= 1e-9:
                        return 0
                    if _re <= 1:
                        _kn = _mach / _re * np.sqrt(_q_interp.gamma * np.pi/2)
                        return 24/_re * (1 + 4.5*_kn)**-1
                    if _re > 1:
                        _kn = _mach / np.sqrt(_re)
                        return 24/_re * (1 + 4.5*_kn)**-1

                case 'henderson':
                    # Henderson model; for all flow regimes
                    # Simplified by ignoring sphere temperature
                    if _re < 1e-9:
                        return 0

                    # For Mach < 1
                    _s = _mach * np.sqrt(_gamma/2)
                    _f1 = 24 * (_re + _s * (5.89688 * np.exp(-0.247 * _re/_s)))**-1
                    _f2 = np.exp(-0.5*_mach/np.sqrt(_re)) * \
                                ((4.5 + 0.38*(0.03*_re + 0.48*np.sqrt(_re))) / (1 + 0.03*_re + 0.48*np.sqrt(_re)) +
                                 0.1*_mach**2 + 0.2*_mach**8)
                    _f3 = (1 - np.exp(-_mach/_re))*0.6*_s
                    _cd1 = _f1 + _f2 + _f3
                    if _mach < 1:
                        return _cd1

                    # For Mach >= 1.75
                    _mach_inf = _mach
                    _re_inf = _re
                    _s_inf = _mach_inf * np.sqrt(_gamma/2)
                    _g1 = 0.9 + 0.34/_mach_inf**2
                    _g2 = 1.86 * np.sqrt(_mach_inf/_re_inf) * (2 + 2/_s_inf**2 + 1.058/_s_inf - 1/_s_inf**4)
                    _g3 = 1 + 1.86 * np.sqrt(_mach_inf/_re_inf)
                    _cd2 = (_g1 + _g2) / _g3
                    if _mach >= 1.75:
                        return _cd2

                    # For 1 <= Mach < 1.75; linear interpolation
                    if 1 <= _mach < 1.75:
                        return _cd1 + 4/3 * (_mach_inf - 1) * (_cd2 - _cd1)

                case 'subramaniam-balachandar':
                    # Model from their new book
                    if _re < 1e-9:
                        return 0

                    if _re < 0.5:
                        # Stokes
                        return 24/_re

                    if _re < 20:
                        # Clift
                        return 24/_re * (1 + 0.1315 * _re**(0.82-0.05*np.log10(_re)))

                    if _re < 800:
                        # Schiller-Naumann
                        return 24/_re * (1 + 0.15 * _re**0.687)

                    if _re < 3e5:
                        # Clift-Gauvin
                        return 24/_re * (1 + 0.15 * _re**0.687 + 0.42/24 * _re * (1 + 4.25e4 * _re**(-1.16))**-1)

                case 'loth':
                    # Loth's model; for all flow regimes
                    if _re < 1e-9:
                        return 0

                    if _re < 45:
                    # Rarefraction dominated domain
                        import math
                        _s = _mach * np.sqrt(_gamma/2)
                        _cd_fm = (1 + 2 * _s**2) * np.exp(-_s**2) / (_s**3 * np.pi**0.5) + \
                                 (4*_s**4 + 4*_s**2 - 1) * math.erf(_s) / (2*_s**4) + 2 * np.pi**0.5 / (3 * _s)
                        _cd_fm_re = _cd_fm / (1 + (_cd_fm/1.63 - 1) * (_re/45)**0.5)
                        _kn = (np.pi * _gamma / 2)**0.5 * _mach / _re
                        _f_kn = (1 + _kn * (2.514 + 0.8 * np.exp(-0.55/_kn)))**-1
                        _cd_kn_re = 24/_re * (1 + 0.15 * _re**0.687) * _f_kn
                        _cd = (_cd_kn_re + _mach**4 * _cd_fm_re) / (1 + _mach**4)
                        return _cd

                    if _re == 45:
                        return 1.63

                    if _re > 45:
                    # compression-dominated regime
                        if _mach <= 1.45:
                            _cm = 5/3 + 2/3 * np.tanh(3 * np.log(_mach + 0.1))
                        if _mach > 1.45:
                            _cm = 2.044 + 0.2 * np.exp(-1.8 * (np.log(_mach/1.5))**2)
                        if _mach <= 0.89:
                            _gm = 1 - 1.525 * _mach**4
                        if _mach > 0.89:
                            _gm = 2e-4 + 8e-4 * np.tanh(12.77 * (_mach - 2.02))
                        _hm = 1 - 0.258 * _cm / (1 + 514 * _gm)
                        _cd = 24/_re * (1 + 0.15 * _re**0.687) * _hm + 0.42 * _cm / (1 + 42000 * _gm / _re**1.16)
                        return _cd

                case 'tedeschi':
                    # Tedeschi's model; for all flow regimes
                    if _re < 1e-9:
                        return 0
                    if _re <= 1:
                        _kn = _mach / _re * np.sqrt(_q_interp.gamma * np.pi/2)
                    else:
                        _kn = _mach / np.sqrt(_re)

                    s = _mach * np.sqrt(_gamma/2)

                    def _solve_k(_k):
                        s_prime = (1 - _k) * s
                        epsilon_prime = 3/8 * (np.pi**2 / s_prime) * (1 + s_prime**2) * s_prime + np.exp(-s_prime**2) /4
                        a1 = 9/4 * 0.15 * 2 * _kn / epsilon_prime * (s * np.pi**0.5 / _kn)**0.687
                        a2 = 1 + 9/4 * 2 * _kn / epsilon_prime
                        return a1 * _k**1.687 + a2 * _k - 1

                    # solve the equation
                    k = fsolve(_solve_k, np.array([0.5]))

                    c = 1 + _re**2 / (_re**2 + 100) * np.e**(-0.225/_mach**2.5)
                    _epsilon_kn = 1.177 + 0.177 * (0.851 * _kn**1.16 - 1) / (0.851 * _kn**1.16 + 1)

                    return 24/_re * k * (1 + 0.15 * (k*_re)**0.687) * c * _epsilon_kn

        def _viscosity(_temperature, law='keyes'):
            """
            Viscosity of air as a function of temperature
            Args:
                _temperature: in Kelvin provided by default in the integration class
                law: 'sutherland' or 'keyes' -- defaults to 'keyes'

            Returns:
                _mu: viscosity of air in kg/m-s

            """
            match law:
                case 'sutherland':
                    # Sutherland's viscosity law
                    # All temperatures must be in kelvin
                    # Formula from cfd-online
                    _c1 = 1.716e-5 * (273.15 + 110.4) / 273.15**1.5
                    _mu = _c1 * _temperature**1.5 * 0.4 / (_temperature + 110.4)
                case 'keyes':
                    # New formula from keyes et al.
                    a0, a, a1 = 1.488, 122.1, 5.0
                    _tau = 1/_temperature
                    _mu = a0 * _temperature**0.5 * 10**-6 / (1 + a * _tau / 10 ** (a1 * _tau))
            return _mu

        match method:
            case 'pRK4':

                def _rk4_step(self, vp, x):
                    """

                    Args:
                        self:

                    Returns:
                        k: ndarray
                            interim RK4 variables

                    """
                    idx = Search(self.interp.idx.grid, x)
                    idx.compute(method='p-space')
                    # For p-space algos; the point-in-domain check was done in search
                    if idx.ppoint is None:
                        return None, None, None
                    interp = Interpolation(self.interp.flow, idx)
                    interp.adaptive = self.interp.adaptive
                    interp.rbf_kernel = self.interp.rbf_kernel
                    interp.compute(method=self.interp.method)
                    q_interp = Variables(interp)
                    # Compute all variables
                    q_interp.compute_mach()
                    uf = q_interp.velocity.reshape(-1)
                    # particle dynamics
                    _rhof = q_interp.density.reshape(-1)
                    dp = diameter
                    rhop = density
                    mu = _viscosity(q_interp.temperature.reshape(-1))
                    # Relative Reynolds Number
                    re = _rhof * np.linalg.norm(vp - uf) * dp / mu
                    _mach = np.linalg.norm(vp - uf) * q_interp.mach.reshape(-1) /\
                            q_interp.velocity_magnitude.reshape(-1)
                    _cd = _drag_constant(re, _q_interp=q_interp, _mach=_mach, _mu=mu, _model=drag_model)
                    _k = -0.75 * _rhof / (rhop * dp)
                    try:
                        _vk = _cd * _k * (vp - uf) * np.linalg.norm(vp - uf) * time_step
                    except TypeError:
                        return None, None, None
                    return _vk, uf, None

                # Start RK4 for p-space
                # For p-space algos; the point-in-domain check was done in search
                x0 = self.interp.idx.ppoint
                if x0 is None:
                    return None, None, None
                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                u0 = q_interp.velocity.reshape(3)
                # Assign velocity to start Rk4 step
                if velocity is None:
                    v0 = u0.copy()
                else:
                    v0 = velocity.copy()
                vk0, uf0, temp = _rk4_step(self, v0, x0)
                xk0 = v0 * time_step
                # Assign fluid velocity when vk is zero
                # Theory: When zero drag particle is massless, hence fluid velocity
                if np.linalg.norm(vk0) == 0:
                    v0 = uf0.copy()
                v1 = v0 + 0.5 * vk0
                x1 = x0 + 0.5 * xk0

                vk1, uf1, temp = _rk4_step(self, v1, x1)
                xk1 = v1 * time_step
                if vk1 is None:
                    return None, None, None
                if np.linalg.norm(vk1) == 0:
                    v0 = uf1.copy()
                v2 = v0 + 0.5 * vk1
                x2 = x0 + 0.5 * xk1
                # Check for mid-RK4 blow-up issue. Happens when Cd and time-step are high
                if np.linalg.norm(x2 - x0) >= 10 * np.linalg.norm(x1-x0) and np.linalg.norm(x2 - x0) >= 1e-12:
                    self.rk4_bool = True
                    return x0, v0, u0

                vk2, uf2, temp = _rk4_step(self, v2, x2)
                xk2 = v2 * time_step
                if vk2 is None:
                    return None, None, None
                if np.linalg.norm(vk2) == 0:
                    v0 = uf2.copy()
                v3 = v0 + vk2
                x3 = x0 + xk2
                # Check for mid-RK4 blow-up issue. Happens when Cd and time-step are high
                if np.linalg.norm(x3 - x0) >= 10 * np.linalg.norm(x1 - x0) and np.linalg.norm(x3 - x0) >= 1e-12:
                    self.rk4_bool = True
                    return x0, v0, u0

                vk3, uf3, temp = _rk4_step(self, v3, x3)
                xk3 = v3 * time_step
                if vk3 is None:
                    return None, None, None
                if np.linalg.norm(vk3) == 0:
                    v0 = uf3.copy()
                v_new = v0 + 1 / 6 * (vk0 + 2 * vk1 + 2 * vk2 + vk3)
                x_new = x0 + 1 / 6 * (xk0 + 2 * xk1 + 2 * xk2 + xk3)
                # Check for mid-RK4 blow-up issue. Happens when Cd and time-step are high
                if np.linalg.norm(x_new - x0) >= 10 * np.linalg.norm(x1 - x0) and np.linalg.norm(x_new - x0) >= 1e-12:
                    self.rk4_bool = True
                    return x0, v0, u0

                vk_new, uf_new, temp = _rk4_step(self, v_new, x_new)
                # make sure the velocity is going down in a compression case
                # This is done to remove the oscillations in the post-shock region
                if vk_new is None:
                    return None, None, None
                if self.interp.method == 'simple_oblique_shock' and np.dot(v_new - v0, uf_new - v_new) < 0:
                    self.rk4_bool = True
                    return x0, v0, u0

                self.ppoint = x_new

                return x_new, v_new, uf_new

            case 'cRK4':

                def _rk4_step(self, c_vp, x):
                    """
                    Returns required data for the RK4 step

                    Args:
                        self:
                        c_vp: ndarray
                            particle velocity in c-space
                        x: ndarray
                            point in c-space

                    Returns:
                        k: ndarray
                            interim RK4 variables

                    """
                    if np.any(x < 0):
                        return None, None, None, None
                    idx = Search(self.interp.idx.grid, x)
                    idx.block = self.interp.idx.block
                    idx.c2p(x)  # This will change the cell attribute
                    # In-domain check is done in search
                    if idx.cpoint is None:
                        self.cpoint = None
                        self.ppoint = None
                        return None, None, None, None
                    interp = Interpolation(self.interp.flow, idx)
                    interp.adaptive = self.interp.adaptive
                    interp.rbf_kernel = self.interp.rbf_kernel
                    interp.compute(method=self.interp.method)
                    _J_inv = interp.J_inv
                    _J = interp.J
                    q_interp = Variables(interp)
                    # Computing mach get all the necessary variables
                    q_interp.compute_mach()
                    p_uf = q_interp.velocity.reshape(-1)
                    c_uf = np.matmul(_J_inv, p_uf)
                    # particle dynamics
                    _rhof = q_interp.density.reshape(-1)
                    dp = diameter
                    rhop = density
                    # Transform particle velocity to p-space
                    p_vp = np.matmul(_J, c_vp)
                    mu = _viscosity(q_interp.temperature.reshape(-1))
                    # Relative Reynolds Number
                    re = _rhof * np.linalg.norm(p_vp - p_uf) * dp / mu
                    _mach = np.linalg.norm(p_vp - p_uf) * q_interp.mach.reshape(-1) /\
                            q_interp.velocity_magnitude.reshape(-1)
                    _cd = _drag_constant(re, _q_interp=q_interp, _mach=_mach, _mu=mu, _model=drag_model)
                    _k = -0.75 * _rhof / (rhop * dp)
                    p_vk = _cd * _k * (p_vp - p_uf) * np.linalg.norm(p_vp - p_uf) * time_step
                    c_vk = np.matmul(_J_inv, p_vk)
                    return c_vk, p_uf, c_uf, p_vp

                # Start RK4 for c-space
                x0 = self.interp.idx.cpoint
                if x0 is None:
                    return None, None, None
                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                p_u0 = q_interp.velocity.reshape(-1)
                c_u0 = np.matmul(self.interp.J_inv, p_u0)
                # Assign velocity to start Rk4 step
                if velocity is None:
                    c_v0 = c_u0.copy()
                else:
                    c_v0 = np.matmul(self.interp.J_inv, velocity)
                vk0, p_u0, c_u0, p_v0 = _rk4_step(self, c_v0, x0)
                # Assign fluid velocity when vk is zero
                # Theory: When zero drag particle is massless, hence fluid velocity
                if np.linalg.norm(vk0) == 0:
                    c_v0 = c_u0.copy()
                c_v1 = c_v0 + 0.5 * vk0
                xk0 = c_v0 * time_step
                x1 = x0 + 0.5 * xk0

                # Integration starts
                vk1, p_u1, c_u1, p_v1 = _rk4_step(self, c_v1, x1)
                # if the residual is none return; exited the domain
                if vk1 is None:
                    return None, None, None
                # if zero; particle is acting like massless particle; low relative reynolds number cases
                if np.linalg.norm(vk1) == 0:
                    c_v0 = c_u1.copy()
                c_v2 = c_v0 + 0.5 * vk1
                xk1 = c_v1 * time_step
                x2 = x0 + 0.5 * xk1
                # Check for mid-RK4 blow up due to residuals
                if np.linalg.norm(x2 - x0) >= 10 * np.linalg.norm(x1-x0):
                    self.rk4_bool = True
                    return x0, p_v0, p_u0

                # Repeat three more times; RK4
                vk2, p_u2, c_u2, p_v2 = _rk4_step(self, c_v2, x2)
                if vk2 is None:
                    return None, None, None
                if np.linalg.norm(vk2) == 0:
                    c_v0 = c_u2.copy()
                c_v3 = c_v0 + vk2
                xk2 = c_v2 * time_step
                x3 = x0 + xk2
                if np.linalg.norm(x3 - x0) >= 10 * np.linalg.norm(x1-x0):
                    self.rk4_bool = True
                    return x0, p_v0, p_u0

                vk3, p_u3, c_u3, p_v3 = _rk4_step(self, c_v3, x3)
                if vk3 is None:
                    return None, None, None
                if np.linalg.norm(vk3) == 0:
                    c_v0 = c_u3.copy()
                c_v_new = c_v0 + 1 / 6 * (vk0 + 2 * vk1 + 2 * vk2 + vk3)
                xk3 = c_v3 * time_step
                x_new = x0 + 1 / 6 * (xk0 + 2 * xk1 + 2 * xk2 + xk3)
                if np.linalg.norm(x_new - x0) >= 10 * np.linalg.norm(x1-x0):
                    self.rk4_bool = True
                    return x0, p_v0, p_u0

                _, p_u_new, c_u_new, p_v_new = _rk4_step(self, c_v_new, x_new)

                self.cpoint = x_new

                return x_new, p_u_new, p_v_new

            case 'unsteady-pRK4':

                def _rk4_step(self, vp, x):
                    """

                    Args:
                        self:

                    Returns:
                        k: ndarray
                            interim RK4 variables

                    """
                    idx = Search(self.interp.idx.grid, x)
                    idx.compute(method='p-space')
                    # For p-space algos; the point-in-domain check was done in search
                    if idx.ppoint is None:
                        return None, None, None
                    interp = Interpolation(self.interp.flow, idx)
                    interp.time = self.interp.time
                    interp.flow_old = self.interp.flow_old
                    interp.adaptive = self.interp.adaptive
                    interp.rbf_kernel = self.interp.rbf_kernel
                    interp.compute(method='unsteady-rbf-p-space')
                    q_interp = Variables(interp)
                    # Compute all variables
                    q_interp.compute_mach()
                    uf = q_interp.velocity.reshape(-1)
                    # particle dynamics
                    _rhof = q_interp.density.reshape(-1)
                    dp = diameter
                    rhop = density
                    mu = _viscosity(q_interp.temperature.reshape(-1), law='sutherland')
                    # Relative Reynolds Number
                    re = _rhof * np.linalg.norm(vp - uf) * dp / mu
                    _mach = np.linalg.norm(vp - uf) * q_interp.mach.reshape(-1) / \
                            q_interp.velocity_magnitude.reshape(-1)
                    _cd = _drag_constant(re, _q_interp=q_interp, _mach=_mach, _mu=mu, _model=drag_model)
                    _k = -0.75 * _rhof / (rhop * dp)
                    try:
                        _vk = _cd * _k * (vp - uf) * np.linalg.norm(vp - uf) * time_step
                    except TypeError:
                        return None, None, None
                    return _vk, uf, None

                # Start RK4 for p-space
                # For p-space algos; the point-in-domain check was done in search
                x0 = self.interp.idx.ppoint
                if x0 is None:
                    return None, None, None
                q_interp = Variables(self.interp)
                q_interp.compute_velocity()
                u0 = q_interp.velocity.reshape(3)
                # Assign velocity to start Rk4 step
                if velocity is None:
                    v0 = u0.copy()
                else:
                    v0 = velocity.copy()
                vk0, uf0, temp = _rk4_step(self, v0, x0)
                xk0 = v0 * time_step
                # Assign fluid velocity when vk is zero
                # Theory: When zero drag particle is massless, hence fluid velocity
                if np.linalg.norm(vk0) == 0:
                    v0 = uf0.copy()
                v1 = v0 + 0.5 * vk0
                x1 = x0 + 0.5 * xk0

                vk1, uf1, temp = _rk4_step(self, v1, x1)
                xk1 = v1 * time_step
                if vk1 is None:
                    return None, None, None
                if np.linalg.norm(vk1) == 0:
                    v0 = uf1.copy()
                v2 = v0 + 0.5 * vk1
                x2 = x0 + 0.5 * xk1
                # Check for mid-RK4 blow-up issue. Happens when Cd and time-step are high
                if np.linalg.norm(x2 - x0) >= 10 * np.linalg.norm(x1 - x0) and np.linalg.norm(x2 - x0) >= 1e-12:
                    self.rk4_bool = True
                    return x0, v0, u0

                vk2, uf2, temp = _rk4_step(self, v2, x2)
                xk2 = v2 * time_step
                if vk2 is None:
                    return None, None, None
                if np.linalg.norm(vk2) == 0:
                    v0 = uf2.copy()
                v3 = v0 + vk2
                x3 = x0 + xk2
                # Check for mid-RK4 blow-up issue. Happens when Cd and time-step are high
                if np.linalg.norm(x3 - x0) >= 10 * np.linalg.norm(x1 - x0) and np.linalg.norm(x3 - x0) >= 1e-12:
                    self.rk4_bool = True
                    return x0, v0, u0

                vk3, uf3, temp = _rk4_step(self, v3, x3)
                xk3 = v3 * time_step
                if vk3 is None:
                    return None, None, None
                if np.linalg.norm(vk3) == 0:
                    v0 = uf3.copy()
                v_new = v0 + 1 / 6 * (vk0 + 2 * vk1 + 2 * vk2 + vk3)
                x_new = x0 + 1 / 6 * (xk0 + 2 * xk1 + 2 * xk2 + xk3)
                # Check for mid-RK4 blow-up issue. Happens when Cd and time-step are high
                if np.linalg.norm(x_new - x0) >= 10 * np.linalg.norm(x1 - x0) and np.linalg.norm(
                        x_new - x0) >= 1e-12:
                    self.rk4_bool = True
                    return x0, v0, u0

                vk_new, uf_new, temp = _rk4_step(self, v_new, x_new)
                # make sure the velocity is going down in a compression case
                # This is done to remove the oscillations in the post-shock region
                if vk_new is None:
                    return None, None, None
                if self.interp.method == 'simple_oblique_shock' and np.dot(v_new - v0, uf_new - v_new) < 0:
                    self.rk4_bool = True
                    return x0, v0, u0

                self.ppoint = x_new

                return x_new, v_new, uf_new



# Use tri-linear interpolation to get data at the given point

import numpy as np
from .variables import Variables


class Interpolation:
    """
    Module to do the cell data interpolation to the given point

    ...

    Attributes
    ----------
    Input:
        flow : .io.plot3dio.FlowIO
            Flow object created from FlowIO
        idx: .streamlines.Search
            Index object created from Search
    Output:
        q : ndarray
            Interpolated data at the given point
        nb, ni, nj, nk: int
            Dimensionless data from flow
        mach, alpha, rey, time: float
            Dimensionless data from flow
        J_inv: ndarray
            Inverse grid metrics at the given point

    Methods
    -------
    compute(method='p-space')
        Interpolates the data onto a given point
        p-space or c-space can be specified

    Example:
        grid = GridIO('../data/plate_data/plate.sp.x')
        flow = FlowIO('../data/plate_data/sol-0000010.q')
        idx = Search(grid, [8.5, 0.5, 0.01])
        interp_data = Interpolation(grid, flow, idx)

        grid.read_grid()
        flow.read_flow()
        idx.compute()
        interp_data.compute()  # method='p-space' is default

    author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
    date: 10-29/2021
    """

    def __init__(self, flow, idx):
        self.flow = flow
        self.idx = idx
        self.q = None
        self.nb = None
        self.ni, self.nj, self.nk = [None] * 3
        self.mach, self.alpha, self.rey, self.time = [None] * 4
        self.J = None
        self.J_inv = None
        self.level = [0, 0, 0]
        self.rbf_kernel = 'thin_plate_spline'
        self.adaptive = None
        self.rbf_epsilon = 1  # Default for RBF interpolation
        self.method = None
        # for unsteady case
        self.flow_old = None
        self.time = []

    def __str__(self):
        doc = "This instance uses " + self.flow.filename + " as the flow file " \
                                                           "to compute properties at " + self.idx.ppoint + "\n"
        return doc

    @staticmethod
    def _shock_cell_check(self):
        # Inspect shock cell and assign nearest interpolation
        i0, j0, k0 = self.idx.cell[0, 0], self.idx.cell[0, 1], self.idx.cell[0, 2]
        i1, j1, k1 = self.idx.cell[1, 0], self.idx.cell[1, 1], self.idx.cell[1, 2]
        # compute velocity, mach
        _var = Variables(self.flow)
        _var.compute_mach()
        # _grad_v = _J_inv * (v1 - v0)
        _grad_v = self.idx.grid.m2[i0, j0, k0, :, self.idx.block] * (
                _var.velocity[i1, j1, k1, :, self.idx.block]
                - _var.velocity[i0, j0, k0, :, self.idx.block]
        )
        # compute norm to get the unit vector
        _grad_v = _grad_v / np.linalg.norm(_grad_v)
        # mach vector -- mach * unit velocity vector
        _mach0 = (_var.mach[i0, j0, k0] * _var.velocity[i0, j0, k0, :, self.idx.block]
                  / _var.velocity_magnitude[i0, j0, k0, self.idx.block])
        _mach1 = (_var.mach[i1, j1, k1] * _var.velocity[i1, j1, k1, :, self.idx.block]
                  / _var.velocity_magnitude[i1, j1, k1, self.idx.block])
        # normal mach vector
        _mach_n0 = np.linalg.norm(np.dot(_mach0, _grad_v))
        _mach_n1 = np.linalg.norm(np.dot(_mach1, _grad_v))

        return _mach_n0, _mach_n1

    def compute(self, method='p-space'):
        """
        Find interpolated plot3d data and grid metrics at a given point
        Args:
            method: Available methods for interpolation
            1. p-space: Tri-linear interpolation in physical co-ordinates
            2. c-space: Tri-linear interpolation in c-space. Must-use for the c-space algo

        Returns:
            q --> Attribute with the interpolated flow data.
            Has same ndim as q attribute from flow object

        """
        # this object is used for integration without changing much of the code
        self.method = method

        # Assign data from q file to keep the format for further computations
        self.nb = self.flow.nb
        self.ni, self.nj, self.nk = self.flow.ni, self.flow.nj, self.flow.nk
        self.mach, self.alpha, self.rey, self.time = self.flow.mach, self.flow.alpha, self.flow.rey, self.flow.time

        # If out of domain return
        if self.idx.ppoint is None and self.idx.cpoint is None:
            print('Cannot run interpolation. The given point is out of domain.\n')
            self.q = None
            return

        # Obtain grid data from nodes found from the search method
        _cell_grd = np.zeros(self.idx.cell.shape)
        _cell_grd = self.idx.grid.grd[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, self.idx.block]

        # Obtain flow data from nodes found
        _cell_q = np.zeros((self.idx.cell.shape[0], 5))
        _cell_q = self.flow.q[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, self.idx.block]

        match method:
            # simple oblique shock -- use nearest neighbor
            case 'simple_oblique_shock':
                if self.idx.cell.shape == (8, 3) and self.idx.info is None:
                    _distance = np.sqrt(np.sum((_cell_grd - self.idx.ppoint) ** 2, axis=1))
                    # nearest neighbor index
                    _nn = np.argmin(_distance)
                    # assign the nearest neighbor to the given point
                    self.q = _cell_q[_nn]
                    self.q = self.q.reshape((1, 1, 1, -1, 1))

                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-12.\n' \
                                    'Interpolation will assign node properties for integration.\n' \
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q[0]  # the first node is the point based on search method
                    self.q = self.q.reshape((1, 1, 1, -1, 1))

            # "Tri"Linear interpolation
            case 'p-space':
                # If the node is found in a cell
                if self.idx.cell.shape == (8, 3) and self.idx.info is None:
                    def _f(self, _m, _n, _arr1, _arr2, _axis, _paxis, _data1, _data2):
                        """
                        Internal function used for linearly interpolating data
                        :param _m: int
                            point-1 for interpolation
                        :param _n: int
                            point-2 for interpolation
                        :param _arr1: ndarray
                            data with point-1
                        :param _arr2: ndarray
                            data with point-2
                        :param _axis: int
                            assign x, y or z axis to consider for interpolation
                        :param _paxis: int
                            assign point's x, y, or z for interpolation
                        :param _data1: ndarray/list
                            data at point-1 to be interpolated. List is provided when moving across axis
                        :param _data2: ndarray/list
                            data at point-2 to be interpolated. List is provided when moving across axis
                        :return: None

                        author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
                        date: 10-29/2021
                        """
                        _a = (_arr2[_n, _axis] - self.idx.ppoint[_paxis]) / (_arr2[_n, _axis] - _arr1[_m, _axis]) * _data1[_m]
                        _b = (self.idx.ppoint[_paxis] - _arr1[_m, _axis]) / (_arr2[_n, _axis] - _arr1[_m, _axis]) * _data2[_n]
                        return _a + _b

                    # 0123 are vertices of quadrilateral
                    # Doing interpolation to x location in self.idx.ppoint
                    # Values in the function help set the linear interpolation in x-direction
                    _cell_grd_01 = _f(self, 0, 1, _cell_grd, _cell_grd, 0, 0, _cell_grd, _cell_grd)
                    _cell_grd_32 = _f(self, 3, 2, _cell_grd, _cell_grd, 0, 0, _cell_grd, _cell_grd)
                    _cell_q_01 = _f(self, 0, 1, _cell_grd, _cell_grd, 0, 0, _cell_q, _cell_q)
                    _cell_q_32 = _f(self, 3, 2, _cell_grd, _cell_grd, 0, 0, _cell_q, _cell_q)

                    # Doing interpolation to the point projection on the face
                    _cell_grd_0123 = _f(self, 1, 1, _cell_grd_01, _cell_grd_32, None, 1,
                                        [None, _cell_grd_01], [None, _cell_grd_32])
                    _cell_q_0123 = _f(self, 1, 1, _cell_grd_01, _cell_grd_32, None, 1,
                                      [None, _cell_q_01], [None, _cell_q_32])

                    # 4567 face
                    # Doing interpolation to x location in self.idx.ppoint
                    # Values in the function help set the linear interpolation in y-direction
                    _cell_grd_45 = _f(self, 4, 5, _cell_grd, _cell_grd, 0, 0, _cell_grd, _cell_grd)
                    _cell_grd_76 = _f(self, 7, 6, _cell_grd, _cell_grd, 0, 0, _cell_grd, _cell_grd)
                    _cell_q_45 = _f(self, 4, 5, _cell_grd, _cell_grd, 0, 0, _cell_q, _cell_q)
                    _cell_q_76 = _f(self, 7, 6, _cell_grd, _cell_grd, 0, 0, _cell_q, _cell_q)

                    # Doing interpolation to the point projection on the face
                    _cell_grd_4567 = _f(self, 1, 1, _cell_grd_45, _cell_grd_76, None, 1, [None, _cell_grd_45],
                                        [None, _cell_grd_76])
                    _cell_q_4567 = _f(self, 1, 1, _cell_grd_45, _cell_grd_76, None, 1, [None, _cell_q_45],
                                      [None, _cell_q_76])

                    # Do the shock cell check
                    if self.idx.cell.shape == (8, 3) and self.idx.info is None:
                        if self.adaptive == 'shock':
                            _mach_n0, _mach_n1 = self._shock_cell_check(self)
                            # if shock is in the cell _mach_n0 > 1 > _mach_n1
                            if _mach_n0 > 1 > _mach_n1:
                                _distance = np.sqrt(np.sum((_cell_grd - self.idx.ppoint) ** 2, axis=1))
                                # nearest neighbor index
                                _nn = np.argmin(_distance)
                                # assign the nearest neighbor to the given point
                                self.q = _cell_q[_nn]
                                self.q = self.q.reshape((1, 1, 1, -1, 1))
                                return
                            else:
                                pass

                    # Doing data interpolation to the given point based on the face points
                    self.q = _f(self, 2, 2, _cell_grd_0123, _cell_grd_4567, None, 2,
                                [None, None, _cell_q_0123], [None, None, _cell_q_4567])
                    self.q = self.q.reshape((1, 1, 1, -1, 1))

                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-12.\n' \
                                    'Interpolation will assign node properties for integration.\n'\
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q[0]  # the first node is the point based on search method
                    self.q = self.q.reshape((1, 1, 1, -1, 1))

            case 'c-space':
                """
                Point from idx is in c-space. Tri-linear interpolation is performed.
                Equation can be found in Sadarjoen et al.
                """

                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-12.\n' \
                                    'Interpolation will assign node properties for integration.\n' \
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q[0]
                    self.q = self.q.reshape((1, 1, 1, -1, 1))
                    return

                # Interpolate m1 and m2 for use in integration; m1 is used in ppath only!
                # m1 -- indicated as J (determinant of m1 is J)
                _cell_J = self.idx.grid.m1[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, :, self.idx.block]
                # m2 -- indicated as J_inv (determinant of m2 is J_inv)
                _cell_J_inv = self.idx.grid.m2[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, :, self.idx.block]
                _alpha, _beta, _gamma = self.idx.cpoint - self.idx.cell[0]

                # Do the shock cell check
                if self.idx.cell.shape == (8, 3) and self.idx.info is None:
                    if self.adaptive == "shock":
                        _mach_n0, _mach_n1 = self._shock_cell_check(self)
                        # if shock is in the cell _mach_n0 > 1 > _mach_n1
                        if _mach_n0 > 1 > _mach_n1:
                            _distance = np.sqrt(np.sum((self.idx.cell - self.idx.cpoint) ** 2, axis=1))
                            # nearest neighbor index
                            _nn = np.argmin(_distance)
                            # assign the nearest neighbor to the given point
                            self.q = _cell_q[_nn]
                            self.J = _cell_J[_nn]
                            self.J_inv = _cell_J_inv[_nn]
                            self.q = self.q.reshape((1, 1, 1, -1, 1))
                            return
                        else:
                            pass

                def _eqn(_alpha, _beta, _gamma, _var):
                    _fun = (1 - _alpha) * (1 - _beta) * (1 - _gamma) * _var[0] + \
                                _alpha  * (1 - _beta) * (1 - _gamma) * _var[1] + \
                                _alpha  *      _beta  * (1 - _gamma) * _var[2] + \
                           (1 - _alpha) *      _beta  * (1 - _gamma) * _var[3] + \
                           (1 - _alpha) * (1 - _beta) *      _gamma  * _var[4] + \
                                _alpha  * (1 - _beta) *      _gamma  * _var[5] + \
                                _alpha  *      _beta *       _gamma  * _var[6] + \
                           (1 - _alpha) *      _beta  *      _gamma  * _var[7]

                    return _fun

                self.q = _eqn(_alpha, _beta, _gamma, _cell_q)
                self.J = _eqn(_alpha, _beta, _gamma, _cell_J)
                self.J_inv = _eqn(_alpha, _beta, _gamma, _cell_J_inv)

                self.q = self.q.reshape((1, 1, 1, -1, 1))

            case 'rbf-p-space':
                """
                Raidal basis function interpolation in physical space
                """
                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-12.\n' \
                                    'Interpolation will assign node properties for integration.\n' \
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q[0]
                    self.q = self.q.reshape((1, 1, 1, -1, 1))
                    return

                from scipy.interpolate import RBFInterpolator as rbf
                # add adjacent cells to the cell list
                # TODO: debug for multiblock case -- currently defaults to single block; if not enough cells in
                #  the block, it default to using rbf
                if np.any(np.array(self.level) > 0):
                    _level_cell = self.idx.cell
                    if (np.any(self.idx.cell[:, 0] + self.level[0] >= self.idx.grid.ni - 1) or
                            np.any(self.idx.cell[:, 0] - self.level[0] <= 0)):
                        pass
                    else:
                        for _i in range(self.level[0]):
                            _level_cell = np.vstack((_level_cell, self.idx.cell[1] + [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[2] + [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[5] + [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[6] + [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[0] - [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[3] - [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[4] - [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[7] - [_i+1, 0, 0]))
                    if (np.any(self.idx.cell[:, 1] + self.level[1] >= self.idx.grid.nj - 1) or
                            np.any(self.idx.cell[:, 1] - self.level[1] <= 0)):
                        pass
                    else:
                        for _i in range(self.level[1]):
                            _level_cell = np.vstack((_level_cell, self.idx.cell[2] + [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[3] + [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[6] + [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[7] + [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[1] - [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[0] - [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[5] - [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[4] - [0, _i+1, 0]))
                    if (np.any(self.idx.cell[:, 2] + self.level[2] >= self.idx.grid.nk - 1) or
                            np.any(self.idx.cell[:, 2] - self.level[2] <= 0)):
                        pass
                    else:
                        for _i in range(self.level[2]):
                            _level_cell = np.vstack((_level_cell, self.idx.cell[4] + [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[5] + [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[6] + [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[7] + [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[0] - [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[1] - [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[2] - [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[3] - [0, 0, _i+1]))
                    # only unique rows to remove singularity issue -- This doesn't happen anymore
                    # Leave it here for now until more testing is done
                    # _level_cell = np.unique(_level_cell, axis=0)
                    # remove rows with negative values -- This doesn't happen anymore
                    # Leave it here for now until more testing is done
                    # _level_cell = _level_cell[~np.any(_level_cell < 0, axis=1)]
                    _cell_grd = self.idx.grid.grd[_level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2],
                                :, self.idx.block]
                    _cell_q = self.flow.q[_level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2],
                                :, self.idx.block]
                    # print('**SUCCESS Interpolating with {} levels'.format(self.level))
                # if -ve values; use the whole grid
                elif np.any(np.array(self.level) < 0):
                    _cell_grd = self.idx.grid.grd[..., self.idx.block].reshape(-1, 3)
                    _cell_q = self.flow.q[..., self.idx.block].reshape(-1, 5)
                _rbf = rbf(_cell_grd, _cell_q, kernel=self.rbf_kernel, epsilon=self.rbf_epsilon)
                self.q = _rbf(np.array(self.idx.ppoint).reshape(1, -1))
                self.q = self.q.reshape((1, 1, 1, -1, 1))

            case 'rbf-c-space':
                """
                Raidal basis function interpolation in c-space
                """
                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-12.\n' \
                                    'Interpolation will assign node properties for integration.\n' \
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q[0]
                    self.q = self.q.reshape((1, 1, 1, -1, 1))
                    return

                # Interpolate m1 and m2 for use in integration; m1 is used in ppath only!
                # m1 -- indicated as J (determinant of m1 is J)
                _cell_J = self.idx.grid.m1[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, :,
                          self.idx.block]
                # m2 -- indicated as J_inv (determinant of m2 is J_inv)
                _cell_J_inv = self.idx.grid.m2[self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, :,
                              self.idx.block]
                # _alpha, _beta, _gamma and reshape for rbf intepolator to work
                _fractions = (self.idx.cpoint - self.idx.cell[0]).reshape(1, -1)

                # Start RBF interpolation
                from scipy.interpolate import RBFInterpolator as rbf
                # add adjacent cells to the cell list
                # TODO: debug for multiblock case -- currently defaults to single block
                _unit_cell = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0],
                                       [0, 1, 0], [0, 0, 1], [1, 0, 1],
                                       [1, 1, 1], [0, 1, 1]])
                if np.any(np.array(self.level) > 0):
                    _level_cell = self.idx.cell
                    if (np.any(self.idx.cell[:, 0] + self.level[0] >= self.idx.grid.ni - 1) or
                            np.any(self.idx.cell[:, 0] - self.level[0] <= 0)):
                        pass
                    else:
                        for _i in range(self.level[0]):
                            _level_cell = np.vstack((_level_cell, self.idx.cell[1] + [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[2] + [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[5] + [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[6] + [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[0] - [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[3] - [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[4] - [_i+1, 0, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[7] - [_i+1, 0, 0]))
                            _unit_cell = np.vstack((_unit_cell, [_i + 2, 0, 0]))
                            _unit_cell = np.vstack((_unit_cell, [_i + 2, 1, 0]))
                            _unit_cell = np.vstack((_unit_cell, [_i + 2, 0, 1]))
                            _unit_cell = np.vstack((_unit_cell, [_i + 2, 1, 1]))
                            _unit_cell = np.vstack((_unit_cell, [-_i -1, 0, 0]))
                            _unit_cell = np.vstack((_unit_cell, [-_i -1, 1, 0]))
                            _unit_cell = np.vstack((_unit_cell, [-_i -1, 0, 1]))
                            _unit_cell = np.vstack((_unit_cell, [-_i -1, 1, 1]))

                    if (np.any(self.idx.cell[:, 1] + self.level[1] >= self.idx.grid.nj - 1) or
                            np.any(self.idx.cell[:, 1] - self.level[1] <= 0)):
                        pass
                    else:
                        for _i in range(self.level[1]):
                            _level_cell = np.vstack((_level_cell, self.idx.cell[2] + [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[3] + [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[6] + [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[7] + [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[0] - [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[1] - [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[4] - [0, _i+1, 0]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[5] - [0, _i+1, 0]))
                            _unit_cell = np.vstack((_unit_cell, [1, _i + 2, 0]))
                            _unit_cell = np.vstack((_unit_cell, [0, _i + 2, 0]))
                            _unit_cell = np.vstack((_unit_cell, [1, _i + 2, 1]))
                            _unit_cell = np.vstack((_unit_cell, [0, _i + 2, 1]))
                            _unit_cell = np.vstack((_unit_cell, [0, -_i - 1, 0]))
                            _unit_cell = np.vstack((_unit_cell, [1, -_i - 1, 0]))
                            _unit_cell = np.vstack((_unit_cell, [0, -_i - 1, 1]))
                            _unit_cell = np.vstack((_unit_cell, [1, -_i - 1, 1]))
                    if (np.any(self.idx.cell[:, 2] + self.level[2] >= self.idx.grid.nk - 1) or
                            np.any(self.idx.cell[:, 2] - self.level[2] <= 0)):
                        pass
                    else:
                        for _i in range(self.level[2]):
                            _level_cell = np.vstack((_level_cell, self.idx.cell[4] + [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[5] + [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[6] + [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[7] + [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[0] - [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[1] - [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[2] - [0, 0, _i+1]))
                            _level_cell = np.vstack((_level_cell, self.idx.cell[3] - [0, 0, _i+1]))
                            _unit_cell = np.vstack((_unit_cell, [0, 0, _i + 2]))
                            _unit_cell = np.vstack((_unit_cell, [1, 0, _i + 2]))
                            _unit_cell = np.vstack((_unit_cell, [1, 1, _i + 2]))
                            _unit_cell = np.vstack((_unit_cell, [0, 1, _i + 2]))
                            _unit_cell = np.vstack((_unit_cell, [0, 0, -_i - 1]))
                            _unit_cell = np.vstack((_unit_cell, [1, 0, -_i - 1]))
                            _unit_cell = np.vstack((_unit_cell, [1, 1, -_i - 1]))
                            _unit_cell = np.vstack((_unit_cell, [0, 1, -_i - 1]))
                    _cell_grd = self.idx.grid.grd[
                        _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, self.idx.block
                    ]
                    _cell_q = self.flow.q[
                        _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, self.idx.block
                    ]
                    _cell_J = self.idx.grid.m1[
                        _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, :, self.idx.block
                    ]
                    _cell_J_inv = self.idx.grid.m2[
                        _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, :, self.idx.block
                    ]
                    # print('**SUCCESS Interpolating with {} levels'.format(self.level))
                _rbf_q = rbf(_unit_cell, _cell_q, kernel=self.rbf_kernel, epsilon=self.rbf_epsilon)
                self.q = _rbf_q(_fractions)
                self.q = self.q.reshape((1, 1, 1, -1, 1))

                _rbf_J = rbf(_unit_cell, _cell_J, kernel=self.rbf_kernel, epsilon=self.rbf_epsilon)
                self.J = _rbf_J(_fractions).reshape(3, 3)
                _rbf_J_inv = rbf(_unit_cell, _cell_J_inv, kernel=self.rbf_kernel, epsilon=self.rbf_epsilon)
                self.J_inv = _rbf_J_inv(_fractions).reshape(3, 3)


            case 'rgi-p-space':
                # implement cublic spline interpolation in p-space
                # _cell_grd and _cell_q are known
                # get _level_grd and _level_q
                # TODO: debug for multiblock case -- currently defaults to single block; if not enough cells in
                #  the block, it defaults to the single cell case
                if np.any(np.array(self.level) > 0):
                    # add cells in i-direction
                    if np.any(self.idx.cell[:, 0] + self.level[0] >= self.idx.grid.ni - 1) or np.any(
                        self.idx.cell[:, 0] - self.level[0] <= 0
                    ):
                        _i_cell = np.unique(self.idx.cell[:, 0])
                        pass
                    else:
                        _i_cell = np.unique(self.idx.cell[:, 0])
                        for _i in range(self.level[0]):
                            _i_cell = np.hstack((_i_cell, self.idx.cell[1, 0] + _i + 1))
                            _i_cell = np.hstack((_i_cell, self.idx.cell[0, 0] - _i - 1))
                    # add nodes in j-direction
                    if np.any(self.idx.cell[:, 1] + self.level[1] >= self.idx.grid.nj - 1) or np.any(
                        self.idx.cell[:, 1] - self.level[1] <= 0
                    ):
                        _j_cell = np.unique(self.idx.cell[:, 1])
                        pass
                    else:
                        _j_cell = np.unique(self.idx.cell[:, 1])
                        for _i in range(self.level[1]):
                            _j_cell = np.hstack((_j_cell, self.idx.cell[2, 1] + _i + 1))
                            _j_cell = np.hstack((_j_cell, self.idx.cell[0, 1] - _i - 1))
                    # add nodes in k-direction
                    if np.any(self.idx.cell[:, 2] + self.level[2] >= self.idx.grid.nk - 1) or np.any(
                        self.idx.cell[:, 2] - self.level[2] <= 0
                    ):
                        _k_cell = np.unique(self.idx.cell[:, 2])
                        pass
                    else:
                        _k_cell = np.unique(self.idx.cell[:, 2])
                        for _i in range(self.level[2]):
                            _k_cell = np.hstack((_k_cell, self.idx.cell[4, 2] + _i + 1))
                            _k_cell = np.hstack((_k_cell, self.idx.cell[0, 2] - _i - 1))
                    # create a meshgrid of all possible combinations
                    _level_cell = np.meshgrid(np.sort(_i_cell), np.sort(_j_cell), np.sort(_k_cell))
                    # re-order back to an array with coordinates
                    _level_cell = np.moveaxis(np.array(_level_cell), 0, _level_cell[0].ndim).reshape(-1, len(_level_cell))
                    _level_cell = np.unique(_level_cell, axis=0)
                    # remove rows with negative values -- This doesn't happen anymore
                    # Leave it here for now until more testing is done
                    _level_cell = _level_cell[~np.any(_level_cell < 0, axis=1)]
                    _cell_grd = self.idx.grid.grd[
                        _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, self.idx.block
                    ]
                    _cell_q = self.flow.q[
                        _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, self.idx.block
                    ]
                # if -ve values; use the whole grid
                elif np.any(np.array(self.level) < 0):
                    _cell_grd = self.idx.grid.grd[..., self.idx.block].reshape(-1, 3)
                    _cell_q = self.flow.q[..., self.idx.block].reshape(-1, 5)
                # if all zeros; use the cell
                elif np.all(np.array(self.level) == 0):
                    _level_cell = np.unique(self.idx.cell, axis=0)
                    _cell_grd = self.idx.grid.grd[
                                _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, self.idx.block]
                    _cell_q = self.flow.q[
                                _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, self.idx.block]

                # Start the RGI
                from scipy.interpolate import RegularGridInterpolator
                # Find the unique values in each direction
                _x = np.unique(_cell_grd[:, 0])
                _y = np.unique(_cell_grd[:, 1])
                _z = np.unique(_cell_grd[:, 2])
                # Set the shape for reshaping q
                _shape = np.array([len(_x), len(_y), len(_z)])

                if self.adaptive =='shock':
                    _mach_n0, _mach_n1 = self._shock_cell_check(self)
                    # if shock is in the cell _mach_n0 > 1 > _mach_n1
                    if _mach_n0 > 1 > _mach_n1:
                        _method = 'nearest'
                    else:
                        if np.all(_shape >= 6):
                            _method = "quintic"
                        elif np.all(_shape >= 4):
                            _method = "cubic"
                        else:
                            _method = "linear"
                # Depending on the shape set the best possible interpolation method
                else:
                    if np.all(_shape >= 6):
                        _method = 'quintic'
                    elif np.all(_shape >= 4):
                        _method = 'cubic'
                    else:
                        _method = 'linear'

                # Create the RGI for each variable
                _rgi_rho = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 0].reshape(_shape), method=_method)
                _rgi_rho_u = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 1].reshape(_shape), method=_method)
                _rgi_rho_v = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 2].reshape(_shape), method=_method)
                _rgi_rho_w = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 3].reshape(_shape), method=_method)
                _rgi_e = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 4].reshape(_shape), method=_method)

                # Find the values at the point and reshape
                self.q = np.array([_rgi_rho(self.idx.ppoint), _rgi_rho_u(self.idx.ppoint),
                                   _rgi_rho_v(self.idx.ppoint), _rgi_rho_w(self.idx.ppoint),
                                   _rgi_e(self.idx.ppoint)])
                self.q = self.q.reshape((1, 1, 1, -1, 1))

            case 'rgi-c-space':
                # implement rgi interpolation in c-space
                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-12.\n' \
                                    'Interpolation will assign node properties for integration.\n' \
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q[0]
                    self.q = self.q.reshape((1, 1, 1, -1, 1))
                    return

                # Interpolate m1 and m2 for use in integration; m1 is used in ppath only!
                # _alpha, _beta, _gamma and reshape for rbf intepolator to work
                _fractions = (self.idx.cpoint - self.idx.cell[0]).reshape(1, -1)

                # add adjacent cells to the cell list
                # TODO: debug for multiblock case -- currently defaults to single block
                _unit_cell = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0],
                                       [0, 1, 0], [0, 0, 1], [1, 0, 1],
                                       [1, 1, 1], [0, 1, 1]])

                # Add cells in _i, _j, _k directions
                if np.any(np.array(self.level) > 0):
                    # add cells in i-direction
                    if np.any(self.idx.cell[:, 0] + self.level[0] >= self.idx.grid.ni - 1) or np.any(
                        self.idx.cell[:, 0] - self.level[0] <= 0
                    ):
                        _i_cell = np.unique(self.idx.cell[:, 0])
                        _i_unit_cell = _unit_cell[:, 0]
                        pass
                    else:
                        _i_cell = np.unique(self.idx.cell[:, 0])
                        _i_unit_cell = _unit_cell[:, 0]
                        for _i in range(self.level[0]):
                            _i_cell = np.hstack((_i_cell, self.idx.cell[1, 0] + _i + 1))
                            _i_cell = np.hstack((_i_cell, self.idx.cell[0, 0] - _i - 1))
                            _i_unit_cell = np.hstack((_i_unit_cell, _unit_cell[:, 0] + _i + 1))
                            _i_unit_cell = np.hstack((_i_unit_cell, _unit_cell[:, 0] - _i - 1))
                    # add nodes in j-direction
                    if np.any(self.idx.cell[:, 1] + self.level[1] >= self.idx.grid.nj - 1) or np.any(
                            self.idx.cell[:, 1] - self.level[1] <= 0
                    ):
                        _j_cell = np.unique(self.idx.cell[:, 1])
                        _j_unit_cell = _unit_cell[:, 1]
                        pass
                    else:
                        _j_cell = np.unique(self.idx.cell[:, 1])
                        _j_unit_cell = _unit_cell[:, 1]
                        for _i in range(self.level[1]):
                            _j_cell = np.hstack((_j_cell, self.idx.cell[2, 1] + _i + 1))
                            _j_cell = np.hstack((_j_cell, self.idx.cell[0, 1] - _i - 1))
                            _j_unit_cell = np.hstack((_j_unit_cell, _unit_cell[:, 1] + _i + 1))
                            _j_unit_cell = np.hstack((_j_unit_cell, _unit_cell[:, 1] - _i - 1))
                    # add nodes in k-direction
                    if np.any(self.idx.cell[:, 2] + self.level[2] >= self.idx.grid.nk - 1) or np.any(
                            self.idx.cell[:, 2] - self.level[2] <= 0
                    ):
                        _k_cell = np.unique(self.idx.cell[:, 2])
                        _k_unit_cell = _unit_cell[:, 2]
                        pass
                    else:
                        _k_cell = np.unique(self.idx.cell[:, 2])
                        _k_unit_cell = _unit_cell[:, 2]
                        for _i in range(self.level[2]):
                            _k_cell = np.hstack((_k_cell, self.idx.cell[4, 2] + _i + 1))
                            _k_cell = np.hstack((_k_cell, self.idx.cell[0, 2] - _i - 1))
                            _k_unit_cell = np.hstack((_k_unit_cell, _unit_cell[:, 2] + _i + 1))
                            _k_unit_cell = np.hstack((_k_unit_cell, _unit_cell[:, 2] - _i - 1))
                    # create a meshgrid of all possible combinations
                    _level_cell = np.meshgrid(np.sort(_i_cell), np.sort(_j_cell), np.sort(_k_cell))
                    _level_unit_cell = np.meshgrid(np.sort(_i_unit_cell), np.sort(_j_unit_cell), np.sort(_k_unit_cell))
                    # re-order back to an array with coordinates
                    _level_cell = np.moveaxis(np.array(_level_cell), 0, _level_cell[0].ndim).reshape(-1, len(_level_cell))
                    _level_cell = np.unique(_level_cell, axis=0)
                    _level_unit_cell = np.moveaxis(np.array(_level_unit_cell), 0, _level_unit_cell[0].ndim).reshape(-1, len(_level_unit_cell))
                    _level_unit_cell = np.unique(_level_unit_cell, axis=0)
                # if all zeros; use the cell
                elif np.all(np.array(self.level) == 0):
                    _level_cell = np.unique(self.idx.cell, axis=0)
                    _level_unit_cell = _unit_cell

                # assign cell values
                _cell_grd = self.idx.grid.grd[
                            _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, self.idx.block]
                _cell_q = self.flow.q[
                            _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, self.idx.block]
                _cell_J = self.idx.grid.m1[
                          _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, :, self.idx.block]
                _cell_J_inv = self.idx.grid.m2[
                            _level_cell[:, 0], _level_cell[:, 1], _level_cell[:, 2], :, :, self.idx.block]

                # Start the RGI
                from scipy.interpolate import RegularGridInterpolator
                # Find the unique values in each direction
                _x = np.unique(_level_unit_cell[:, 0])
                _y = np.unique(_level_unit_cell[:, 1])
                _z = np.unique(_level_unit_cell[:, 2])
                # Set the shape for reshaping q
                _shape = np.array([len(_x), len(_y), len(_z)])

                if self.adaptive =='shock':
                    _mach_n0, _mach_n1 = self._shock_cell_check(self)
                    # if shock is in the cell _mach_n0 > 1 > _mach_n1
                    if _mach_n0 > 1 > _mach_n1:
                        _method = 'nearest'
                    else:
                        if np.all(_shape >= 6):
                            _method = "quintic"
                        elif np.all(_shape >= 4):
                            _method = "cubic"
                        else:
                            _method = "linear"
                # Depending on the shape set the best possible interpolation method
                else:
                    if np.all(_shape >= 6):
                        _method = 'quintic'
                    elif np.all(_shape >= 4):
                        _method = 'cubic'
                    else:
                        _method = 'linear'

                # Create the RGI for each variable
                _rgi_rho = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 0].reshape(_shape), method=_method)
                _rgi_rho_u = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 1].reshape(_shape), method=_method)
                _rgi_rho_v = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 2].reshape(_shape), method=_method)
                _rgi_rho_w = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 3].reshape(_shape), method=_method)
                _rgi_e = RegularGridInterpolator((_x, _y, _z), _cell_q[:, 4].reshape(_shape), method=_method)

                # Find the values at the point and reshape
                self.q = np.array([_rgi_rho(_fractions), _rgi_rho_u(_fractions),
                                      _rgi_rho_v(_fractions), _rgi_rho_w(_fractions),
                                      _rgi_e(_fractions)])
                self.q = self.q.reshape((1, 1, 1, -1, 1))

                _rgi_J00 = RegularGridInterpolator((_x, _y, _z), _cell_J[:, 0, 0].reshape(_shape), method=_method)
                _rgi_J01 = RegularGridInterpolator((_x, _y, _z), _cell_J[:, 0, 1].reshape(_shape), method=_method)
                _rgi_J02 = RegularGridInterpolator((_x, _y, _z), _cell_J[:, 0, 2].reshape(_shape), method=_method)
                _rgi_J10 = RegularGridInterpolator((_x, _y, _z), _cell_J[:, 1, 0].reshape(_shape), method=_method)
                _rgi_J11 = RegularGridInterpolator((_x, _y, _z), _cell_J[:, 1, 1].reshape(_shape), method=_method)
                _rgi_J12 = RegularGridInterpolator((_x, _y, _z), _cell_J[:, 1, 2].reshape(_shape), method=_method)
                _rgi_J20 = RegularGridInterpolator((_x, _y, _z), _cell_J[:, 2, 0].reshape(_shape), method=_method)
                _rgi_J21 = RegularGridInterpolator((_x, _y, _z), _cell_J[:, 2, 1].reshape(_shape), method=_method)
                _rgi_J22 = RegularGridInterpolator((_x, _y, _z), _cell_J[:, 2, 2].reshape(_shape), method=_method)
                _J00, _J01, _J02 = _rgi_J00(_fractions), _rgi_J01(_fractions), _rgi_J02(_fractions)
                _J10, _J11, _J12 = _rgi_J10(_fractions), _rgi_J11(_fractions), _rgi_J12(_fractions)
                _J20, _J21, _J22 = _rgi_J20(_fractions), _rgi_J21(_fractions), _rgi_J22(_fractions)
                self.J = np.array([[_J00, _J01, _J02], [_J10, _J11, _J12], [_J20, _J21, _J22]]).reshape(3, 3)

                _rgi_J00_inv = RegularGridInterpolator((_x, _y, _z), _cell_J_inv[:, 0, 0].reshape(_shape), method=_method)
                _rgi_J01_inv = RegularGridInterpolator((_x, _y, _z), _cell_J_inv[:, 0, 1].reshape(_shape), method=_method)
                _rgi_J02_inv = RegularGridInterpolator((_x, _y, _z), _cell_J_inv[:, 0, 2].reshape(_shape), method=_method)
                _rgi_J10_inv = RegularGridInterpolator((_x, _y, _z), _cell_J_inv[:, 1, 0].reshape(_shape), method=_method)
                _rgi_J11_inv = RegularGridInterpolator((_x, _y, _z), _cell_J_inv[:, 1, 1].reshape(_shape), method=_method)
                _rgi_J12_inv = RegularGridInterpolator((_x, _y, _z), _cell_J_inv[:, 1, 2].reshape(_shape), method=_method)
                _rgi_J20_inv = RegularGridInterpolator((_x, _y, _z), _cell_J_inv[:, 2, 0].reshape(_shape), method=_method)
                _rgi_J21_inv = RegularGridInterpolator((_x, _y, _z), _cell_J_inv[:, 2, 1].reshape(_shape), method=_method)
                _rgi_J22_inv = RegularGridInterpolator((_x, _y, _z), _cell_J_inv[:, 2, 2].reshape(_shape), method=_method)
                _J00_inv, _J01_inv, _J02_inv = _rgi_J00_inv(_fractions), _rgi_J01_inv(_fractions), _rgi_J02_inv(_fractions)
                _J10_inv, _J11_inv, _J12_inv = _rgi_J10_inv(_fractions), _rgi_J11_inv(_fractions), _rgi_J12_inv(_fractions)
                _J20_inv, _J21_inv, _J22_inv = _rgi_J20_inv(_fractions), _rgi_J21_inv(_fractions), _rgi_J22_inv(_fractions)
                self.J_inv = np.array([[_J00_inv, _J01_inv, _J02_inv],
                                      [_J10_inv, _J11_inv, _J12_inv],
                                      [_J20_inv, _J21_inv, _J22_inv]]).reshape(3, 3)

            # TODO: Thorough testing needed
            case 'unsteady-rbf-p-space':
                """
                Raidal basis function interpolation in physical space for unsteady problems
                """
                # if the point is node return node data
                if self.idx.info == 'Given point is a node in the domain with a tol of 1e-12.\n' \
                                    'Interpolation will assign node properties for integration.\n' \
                                    'Index of the node will be returned by cell attribute\n':
                    self.q = _cell_q[0]
                    self.q = self.q.reshape((1, 1, 1, -1, 1))
                    return

                from scipy.interpolate import RBFInterpolator as rbf
                _rbf = rbf(_cell_grd, _cell_q)
                self.q = _rbf(np.array(self.idx.ppoint).reshape(1, -1))
                self.q = self.q.reshape((1, 1, 1, -1, 1))
                # Equation is linear interpolation between time steps in unsteady data
                # use of try-except is to avoid error in the first time step
                # instead of keeping track of it in every step
                try:
                    _tau = (np.sum(self.time) - self.flow_old.time) / (self.flow.time - self.flow_old.time)
                    _cell_q_old = self.flow_old.q[
                                  self.idx.cell[:, 0], self.idx.cell[:, 1], self.idx.cell[:, 2], :, self.idx.block
                                  ]
                    _rbf_old = rbf(_cell_grd, _cell_q_old)
                    _q_old = _rbf_old(np.array(self.idx.ppoint).reshape(1, -1))
                    _q_old = _q_old.reshape((1, 1, 1, -1, 1))
                    self.q = _tau * self.q + (1 - _tau) * _q_old
                except AttributeError:
                    return



# Creates 3d IA from grid and flow data
# Spawns particles on a distribution and returns their locations
import numpy as np
from multiprocessing import cpu_count, Pool
from ..function.variables import Variables
from ..function.search import Search
from ..function.interpolation import Interpolation
from ..function.integration import Integration
import tqdm
rng = np.random.default_rng()


class Particle:
    """
    Class holds details for particles used in PIV experiment
    ---
    User has to provide all the information
    """

    def __init__(self):
        self.distribution = "gaussian"
        self.min_dia = None
        self.max_dia = None
        self.mean_dia = None
        self.std_dia = None
        self.density = None
        self.n_concentration = None
        self.particle_field = None

    def compute_distribution(self):
        """
        Run this method to return a distribution of particle diameters
        :return: numpy.ndarray
        A 1d array of particle diameters
        """
        if self.distribution == "gaussian":
            print("When Gaussian distribution is used,"
                  " the particle statistics are computed using mean and std diameters\n"
                  "Particle min and max are cutoffs for the distribution")
            self.particle_field = rng.normal(self.mean_dia, self.std_dia, int(self.n_concentration))
            self.particle_field = np.clip(self.particle_field, self.min_dia, self.max_dia)
            rng.shuffle(self.particle_field)
            return

        # TODO: Add Uniform distribution

    pass


# TODO: Add laser pulse for generating second snap
class LaserSheet:
    """
    Laser sheet information to pass into CreateParticles
    ---
    User input class
    """

    def __init__(self, grid):
        self.grid = grid
        self.distribution = "gaussian"
        self.position = None
        self.thickness = None
        self.width = None
        self.pulse_time = None

        print("The laser sheet will reside inside the grid. "
              "Carefully assign position and thickness using parameters below")
        print(f"position and thickness should be between {self.grid.grd_min[:, 2]} and {self.grid.grd_max[:, 2]}")

    def compute_bounds(self):
        self.width = np.array([self.position - self.thickness / 2, self.position + self.thickness / 2])

    pass


class CreateParticles:
    """
    Module to create 3d Interrogation Area (IA) from grid and flow data
    AND
    Spawns particles based on the given distribution
    """

    def __init__(self, grid, flow, particle, laser_sheet, ia_bounds: list[float, float, float, float]):
        self.grid = grid
        self.flow = flow
        self.particle = particle
        self.laser_sheet = laser_sheet
        # x_min, x_max, y_min, y_max --> ia_bounds
        self.ia_bounds = ia_bounds
        # percent of particles in-plane; rest will be divided equally above and below the ia_plane
        self.in_plane = None
        # locations is an n x 4 array; [x, y, z, diameter]
        self.locations = None
        self.locations2 = []
        self._failed_ids = []
        print(f"ia_bounds should be with in:\n"
              f"In x-direction: {self.grid.grd_min[:, 0]} and {self.grid.grd_max[:, 0]}\n"
              f"In y-direction: {self.grid.grd_min[:, 1]} and {self.grid.grd_max[:, 1]}\n")

    def compute_locations(self):
        # Uniform distribution
        # In-plane points
        _particles_in_plane = int(self.in_plane * self.particle.n_concentration * 0.01)
        _x_loc = rng.uniform(self.ia_bounds[0], self.ia_bounds[1], _particles_in_plane)
        _y_loc = rng.uniform(self.ia_bounds[2], self.ia_bounds[3], _particles_in_plane)
        _z_loc = np.repeat(self.laser_sheet.position, _particles_in_plane)
        self.locations = np.vstack((_x_loc, _y_loc, _z_loc, self.particle.particle_field[:_particles_in_plane])).T

        # Off-plane locations - randomize z
        _particles_off_plane = int(self.particle.n_concentration - _particles_in_plane)
        _x_loc = rng.uniform(self.ia_bounds[0], self.ia_bounds[1], _particles_off_plane)
        _y_loc = rng.uniform(self.ia_bounds[2], self.ia_bounds[3], _particles_off_plane)
        _z_loc = rng.uniform(self.laser_sheet.width[0], self.laser_sheet.width[1], _particles_off_plane)
        self.locations = np.concatenate((self.locations,
                                         np.vstack((_x_loc, _y_loc, _z_loc,
                                                    self.particle.particle_field[_particles_in_plane:])).T), axis=0)

        return

    def _process(self, _location, _task_id):
        try:
            _x, _y, _z, _d = _location
            _idx = Search(self.grid, [_x, _y, _z])
            _idx.compute(method='p-space')

            _interp = Interpolation(self.flow, _idx)
            _interp.compute(method='p-space')

            _var = Variables(_interp)
            _var.compute_velocity()

            _intg = Integration(_interp)
            _new_loc, _ = _intg.compute(method='pRK4', time_step=self.laser_sheet.pulse_time)
            if _new_loc is None:
                # make sure the point moves out of the grid for second image
                _new_loc = np.array((_x, _y, _z)) + self.laser_sheet.pulse_time * _var.velocity.reshape(3)
            return np.hstack((_new_loc, _d))
        except:
            # delete the particle from self.locations
            self._failed_ids.append(_task_id)
            print(f"***Error in task {_task_id}***")
            return None

    def compute_locations2(self):
        """
        Will integrate particles to new locations based on
        Laser pulse time and velocities at their locations
        :return:
        """

        # setup parameters for multiprocessing
        _tasks = np.arange(len(self.locations))
        n = max(1, cpu_count() - 1)
        pool = Pool(n)
        self.locations2 = pool.starmap(self._process, zip(self.locations, _tasks))
        pool.close()
        pool.join()

        # delete failed tasks
        self.locations = np.delete(self.locations, self._failed_ids, axis=0)

        self.locations2 = np.array(self.locations2)
        self.locations2 = np.delete(self.locations2, self._failed_ids, axis=0)
        print(f"Total number of particles as per locations: {len(self.locations)}")
        print(f"Total number of particles as per locations2: {len(self.locations2)}")
        print(f"Failed number of particles: {len(self._failed_ids)}")

        return

    def compute_locations2_serial(self):
        """
        Will integrate particles to new locations based on
        Laser pulse time and velocities at their locations
        :return:
        """

        # for loop for serial computation. Track using tqdm computing locations...
        for _i, _j in enumerate(tqdm.tqdm(self.locations, desc="Computing locations for second image")):
            self.locations2.append(self._process(_j, _i))

        # delete failed tasks
        self.locations = np.delete(self.locations, self._failed_ids, axis=0)

        self.locations2 = np.array(self.locations2)
        self.locations2 = np.delete(self.locations2, self._failed_ids, axis=0)
        print(f"Total number of particles as per locations: {len(self.locations)}")
        print(f"Total number of particles as per locations2: {len(self.locations2)}")
        print(f"Failed number of particles: {len(self._failed_ids)}")

        return

    pass



# import all sypiv modules
from .create_particles import Particle, CreateParticles, LaserSheet
from .ccd_projection import CCDProjection
from .intensity import Intensity
from .image_gen import ImageGen



# Generate first and second images imitating PIV

import numpy as np
import matplotlib.pyplot as plt


class ImageGen:
    """
    Class to generate images
    """

    def __init__(self, intensity):
        self.intensity = intensity
        self.fig = None
        self.ax = None
        self.snap_num = None

    def snap(self, snap_num=0):
        """
        Function to generate first snap from the intensity data
        :return:
        """
        self.snap_num = snap_num
        # Specified in inches --> python default
        xsize = self.intensity.projection.xres / self.intensity.projection.dpi
        ysize = self.intensity.projection.yres / self.intensity.projection.dpi
        self.fig = plt.figure(figsize=[xsize, ysize], dpi=self.intensity.projection.dpi)
        ax = plt.axes([0.0, 0.0, 1.0, 1.0])
        ax.imshow(self.intensity.values, cmap='gray', origin='lower')
        ax.axis('tight')
        # remove black border
        ax.axis('off')
        # plt.show()

        return

    def save_snap(self, fname=None):
        if fname is None:
            fname = "snap_" + str(self.snap_num) + ".tif"
        self.fig.savefig(fname=fname)
        # close the figure
        plt.close(self.fig)

        return

    def check_data(self, snap_num=0):
        """
        Function to generate images during the image gen process
        1. Create Particles in physical space
        2. Projections of particles onto CCD
        3. Contour plot of intensity
        4. Image of the particle snaps - This can be saved!
        :return:
        """
        print("Data at various steps during image generation will be plotted...")

        # Plot data after creating particles
        # Use the percent provided to count number of particles
        plt.figure()
        particles = self.intensity.projection.particles
        _in_plane = int(particles.particle.n_concentration *
                        particles.in_plane * 0.01)
        # plot in-plane particle locations
        plt.scatter(particles.locations[:_in_plane, 0], particles.locations[:_in_plane, 1],
                    s=10 * particles.locations[:_in_plane, 3] / particles.particle.min_dia, c='g', label='in-plane')
        # plot out-of-plane locations
        plt.scatter(particles.locations[_in_plane:, 0], particles.locations[_in_plane:, 1],
                    s=10 * particles.locations[_in_plane:, 3] / particles.particle.min_dia, c='r', label='out-of-plane')
        plt.title('Particles created on the given IA')
        plt.legend()

        # Plot data after particle projection
        plt.figure()
        proj = self.intensity.projection
        plt.scatter(proj.projections[:_in_plane, 0], proj.projections[:_in_plane, 1], c='g',
                    s=10 * proj.projections[:_in_plane, 2] / particles.particle.min_dia, label='in-plane')
        # plot out-of-plane locations
        plt.scatter(proj.projections[_in_plane:, 0], proj.projections[_in_plane:, 1], c='r',
                    s=10 * proj.projections[_in_plane:, 2] / particles.particle.min_dia, label='out-of-plane')
        plt.title('Projected data; shown in pixels; particle sizes are representative;\n data is mirrored')
        plt.legend()

        # Plot data in contours
        xsize = proj.xres / self.intensity.projection.dpi
        ysize = proj.yres / self.intensity.projection.dpi
        x = np.linspace(-proj.xres / 2, proj.xres / 2, proj.xres)
        y = np.linspace(-proj.yres / 2, proj.yres / 2, proj.yres)
        fig = plt.figure(figsize=[xsize, ysize], dpi=proj.dpi)
        ax = plt.axes([0.0, 0.0, 1.0, 1.0], xlim=(min(x), max(x)), ylim=(min(y), max(y)))
        ax.contourf(x, y, self.intensity.values, cmap='hot')
        ax.set_title('Contour plot of intensities')

        # Show the image
        self.snap(snap_num=snap_num)
        plt.show()

        return





# Generate intensity field at particle particles
from scipy.special import erf
import numpy as np
from multiprocessing import Pool
from multiprocessing import cpu_count
import dask.array as da
import tqdm


# import sys
# resource.setrlimit(resource.RLIMIT_STACK, (2**29, -1))
# sys.setrecursionlimit(10**6)


class Intensity:
    """
    Parameters
    ----------
    cache : tuple
            (radiusx, radiusy, xp, yp, sx, sy, frx, fry)
            Has all the variables needed for compute function.
            Created to clean up the code
    projection : object from CCDProjection class
            Contains particle location data in pixels
            Contains xres, yres, and dpi data

    Returns
    -------
    intensity: function
            Filled with data from cache; a function of particle locations
    values: numpy.ndarray
            Intensity values for the final image

    By: Dilip Kalagotla ~ kal @ dilip.kalagotla@gmail.com
    Date created: Sat August 6, 2022
    """

    def __init__(self, cache, projection):
        self.cache = cache
        self.projection = projection
        self.intensity = None
        self.values = None

    def setup(self, x, y, ls_thickness, ls_position, dia_x, dia_y, xp, yp, sx, sy, frx, fry, s, q, z_physical, _task):
        """
        cache = (radiusx, radiusy, xp, yp, sx, sy, frx, fry)
        I = intensityField(cache)

        Parameters
        ----------

        Returns
        -------
        intensity : function
            Intensity field as a function of particle particles.

        By: Dilip Kalagotla ~ kal @ dilip.kalagotla@gmail.com
        Date created: Mon May 17 11:00:56 2021

        """
        # start time
        # start = time.perf_counter()

        # compute intensity for each particle location based on Gaussian distribution
        # q is the efficiency factor with which particles scatter light
        # s is the shape factor; 2 --> Gaussian, 10^4 --> uniform

        self.intensity = (q *
                          np.exp(-1 / np.sqrt(2 * np.pi) *
                                 abs(2 * (z_physical - ls_position) ** 2 / ls_thickness ** 2) ** s) *
                          (np.pi / 8 * dia_x * dia_y * sx * sy *
                           (erf((x - xp + 0.5 * frx) / (sx * 2 ** 0.5)) -
                            erf((x - xp - 0.5 * frx) / (sx * 2 ** 0.5))) *
                           (erf((y - yp + 0.5 * fry) / (sy * 2 ** 0.5)) -
                            erf((y - yp - 0.5 * fry) / (sy * 2 ** 0.5)))))
        # end time
        # end = time.perf_counter()

        # print(f'Done computing intensity field for {_task}/{len(self.cache[0])} particles in {end - start} seconds.'
        #       f'process number is {os.getpid()}')

        return self.intensity

    def compute(self, chunksize=5096):
        # Using multiprocessing to compute relative intensity field
        (dia_x, dia_y, xp, yp, sx, sy, frx, fry, s, q, z_physical) = self.cache
        intensity = np.zeros((self.projection.yres, self.projection.xres))
        # create meshgrid for x and y
        x = np.linspace(-self.projection.xres / 2, self.projection.xres / 2, self.projection.xres)
        y = np.linspace(-self.projection.yres / 2, self.projection.yres / 2, self.projection.yres)
        x, y = np.meshgrid(x, y)

        # laser sheet thickness
        ls_thickness = self.projection.particles.laser_sheet.thickness
        ls_position = self.projection.particles.laser_sheet.position

        i = 0
        j = chunksize
        pbar = tqdm.tqdm(total=len(xp), desc="Computing intensity field for particles", position=0, leave=True,
                         colour='green')
        while j <= len(xp):
            n = max(1, cpu_count() - 1)
            pool = Pool(n)
            n_particles = len(dia_x[i:j])
            # x, y, ls_thickness, ls_position,
            # dia_x[i], dia_y[i], xp[i], yp[i], sx, sy, frx, fry, s, q, z_physical[i], i
            itemp = pool.starmap(self.setup, zip([x]*n_particles, [y]*n_particles,
                                                 np.repeat(ls_thickness, n_particles),
                                                 np.repeat(ls_position, n_particles),
                                                 dia_x[i:j], dia_y[i:j], xp[i:j], yp[i:j],
                                                 np.repeat(sx, n_particles), np.repeat(sy, n_particles),
                                                 np.repeat(frx, n_particles), np.repeat(fry, n_particles),
                                                 np.repeat(s, n_particles), np.repeat(q, n_particles), z_physical[i:j],
                                                 np.arange(n_particles)), chunksize=n_particles // n)
            pool.close()
            pool.join()

            intensity += np.sum(itemp, axis=0)
            i = j
            j += chunksize
            # print(f"Done with {i} particles out of {len(xp)}")
            pbar.update(chunksize)

        # compute the remaining particles
        n = max(1, cpu_count() - 1)
        pool = Pool(n)
        n_particles = len(dia_x[i:])
        itemp = pool.starmap(self.setup, zip([x]*n_particles, [y]*n_particles,
                                             np.repeat(ls_thickness, n_particles),
                                             np.repeat(ls_position, n_particles),
                                             dia_x[i:], dia_y[i:], xp[i:], yp[i:],
                                             np.repeat(sx, n_particles), np.repeat(sy, n_particles),
                                             np.repeat(frx, n_particles), np.repeat(fry, n_particles),
                                             np.repeat(s, n_particles), np.repeat(q, n_particles), z_physical[i:],
                                             np.arange(n_particles)))
        pool.close()
        pool.join()
        # print(f"Done with {len(xp)} particles out of {len(xp)}")
        pbar.update(len(xp) - i)
        pbar.close()

        intensity += np.sum(itemp, axis=0)

        # Average intensity field
        intensity = intensity / len(xp)

        # Normalize intensity field to rbg values
        if np.max(intensity) != 0:
            intensity = intensity / np.max(intensity) * 255
        print('Done computing intensity field')

        self.values = intensity

        return self.values

    def compute_serial(self):
        (dia_x, dia_y, xp, yp, sx, sy, frx, fry, s, q, z_physical) = self.cache
        intensity = np.zeros((self.projection.yres, self.projection.xres))
        x = np.linspace(-self.projection.xres / 2, self.projection.xres / 2, self.projection.xres)
        y = np.linspace(-self.projection.yres / 2, self.projection.yres / 2, self.projection.yres)
        x, y = np.meshgrid(x, y)

        # laser sheet thickness
        ls_thickness = self.projection.particles.laser_sheet.thickness
        ls_position = self.projection.particles.laser_sheet.position

        for i in tqdm.tqdm(range(len(xp)), desc="Computing intensity field for particles"):
            intensity += self.setup(x, y, ls_thickness, ls_position,
                                    dia_x[i], dia_y[i], xp[i], yp[i], sx, sy, frx, fry, s, q, z_physical[i], i)

        # Average intensity field
        intensity = intensity / len(xp)

        # Normalize intensity field to rbg values
        if np.max(intensity) != 0:
            intensity = intensity / np.max(intensity) * 255
        print('Done computing intensity field')

        self.values = intensity

        return self.values

    def compute_dask(self):
        """
        Ideal for computing intensity field for higher resolution images
        :return:
        """
        (dia_x, dia_y, xp, yp, sx, sy, frx, fry, s, q, z_physical) = self.cache

        # convert arrays to dask arrays
        dia_x = da.from_array(dia_x, chunks=1000)
        dia_y = da.from_array(dia_y, chunks=1000)
        xp = da.from_array(xp, chunks=1000)
        yp = da.from_array(yp, chunks=1000)
        z_physical = da.from_array(z_physical, chunks=1000)
        intensity = np.zeros((self.projection.yres, self.projection.xres))
        # holder for intensity field -- saves time during addition
        da_intensity = da.zeros((self.projection.yres, self.projection.xres), chunks=(1000, 1000))

        # create meshgrid for x and y
        x = da.linspace(-self.projection.xres / 2, self.projection.xres / 2, self.projection.xres)
        y = da.linspace(-self.projection.yres / 2, self.projection.yres / 2, self.projection.yres)
        x, y = da.meshgrid(x, y)

        # laser sheet thickness
        ls_thickness = self.projection.particles.laser_sheet.thickness
        ls_position = self.projection.particles.laser_sheet.position

        # compute intensity for each particle location based on Gaussian distribution
        # q is the efficiency factor with which particles scatter light
        # s is the shape factor; 2 --> Gaussian, 10^4 --> uniform

        # 200 particles at a time to avoid recursion limit
        for j in tqdm.tqdm(range(0, len(xp), 200), desc=f"Computing intensity field for particles in chunks of 200",
                           position=0, leave=True, colour='green'):
            for i in tqdm.tqdm(range(j, j + 200 if j + 200 < len(xp) else len(xp)),
                               desc=f"Setting up intensity field for particles {j} to"
                                    f" {j + 200 if j + 200 < len(xp) else len(xp)}",
                               position=1, leave=False, colour='blue'):
                da_intensity = (da_intensity +
                                (q *
                                 da.exp(-1 / da.sqrt(2 * np.pi) *
                                        abs(2 * (z_physical[i] - ls_position) ** 2 / ls_thickness ** 2) ** s) *
                                 (np.pi / 8 * dia_x[i] * dia_y[i] * sx * sy *
                                  (erf((x - xp[i] + 0.5 * frx) / (sx * 2 ** 0.5)) -
                                   erf((x - xp[i] - 0.5 * frx) / (sx * 2 ** 0.5))) *
                                  (erf((y - yp[i] + 0.5 * fry) / (sy * 2 ** 0.5)) -
                                   erf((y - yp[i] - 0.5 * fry) / (sy * 2 ** 0.5)))))
                                )
            da_intensity = da_intensity.compute()
            intensity = intensity + da_intensity
            da_intensity = da.zeros((self.projection.yres, self.projection.xres), chunks=(1000, 1000))

        # Average intensity field
        intensity = intensity / len(xp)

        # Normalize intensity field to rbg values
        if np.max(intensity) != 0:
            intensity = intensity / np.max(intensity) * 255
        print('Done computing intensity field')

        self.values = intensity

        return self.values



import unittest
import numpy as np


class TestImageGen(unittest.TestCase):
    def test_image_gen(self):
        from src.sypivlib.function.dataio import GridIO, FlowIO
        from src.sypivlib.sypiv.create_particles import Particle, LaserSheet, CreateParticles
        from src.sypivlib.sypiv.ccd_projection import CCDProjection
        from src.sypivlib.sypiv.intensity import Intensity
        from src.sypivlib.sypiv.image_gen import ImageGen

        # Read-in the grid and flow file
        grid = GridIO('../data/shocks/interpolated_data/mgrd_to_p3d.x')
        grid.read_grid()
        grid.compute_metrics()
        flow = FlowIO('../data/shocks/interpolated_data/mgrd_to_p3d_particle.q')
        flow.read_flow()

        # Set particle data
        p = Particle()
        p.min_dia = 1940e-9  # m
        p.max_dia = 1940e-9  # m
        p.mean_dia = 1940e-9  # m
        p.std_dia = 0  # m
        p.density = 810  # kg/m3
        p.n_concentration = 2
        p.compute_distribution()

        # Read-in the laser sheet
        laser = LaserSheet(grid)
        # z-location
        laser.position = 0.00025  # in m
        laser.thickness = 0.0001  # in m (Data obtained from LaVision)
        laser.pulse_time = 1e-9
        laser.compute_bounds()

        # path to save files
        path = '../data/shocks/interpolated_data/particle_snaps/'

        for i in range(1):
            # Create particle locations array
            ia_bounds = [None, None, None, None]
            loc = CreateParticles(grid, flow, p, laser, ia_bounds)
            # x_min, x_max, y_min, y_max --> ia_bounds
            loc.ia_bounds = [0.0016, 0.0025, 0.0002, 0.0004]  # in m
            loc.in_plane = 100
            loc.compute_locations()
            # To adjust for the test case; set locations manually
            # Two locations out-of-plane case
            # loc.locations = np.array([[0.0016+0.000225, 0.0003, 0.00025, 1940e-9],
            #                           [0.0025-0.000225, 0.0003, 0.00025, 1840e-9]])
            # one location center of the plane
            loc.locations = np.array([[0.0025-0.000225*2, 0.0003, 0.00025, p.mean_dia]])
            loc.compute_locations2()

            # Create particle projections (Simulating data from EUROPIV)
            proj = CCDProjection(loc)
            proj.dpi = 72
            proj.xres = 16
            proj.yres = 16
            # Set distance based on similar triangles relationship
            proj.d_ccd = proj.xres * 25.4e-3 / proj.dpi  # in m
            proj.d_ia = 0.0009  # in m; ia_bounds (max - min)
            proj.compute()

            # (radiusx, radiusy, xp, yp, sx, sy, frx, fry, s, q, z_physical)
            cache = (proj.projections[:, 2], proj.projections[:, 2],
                     proj.projections[:, 0], proj.projections[:, 1],
                     2.0, 2.0, 1.0, 1.0,
                     2, 1, loc.locations[:, 2])  # 2 is gaussian profile, 1 is reflectivity factor, z_physical
            intensity = Intensity(cache, proj)
            intensity.compute()

            snap = ImageGen(intensity)
            snap.snap(snap_num=1)
            # snap.save_snap(fname=path + str(i) + '_1.tif')
            snap.check_data(snap_num=1)
            print('Done with image 1 for pair number ' + str(i) + '\n')

            cache2 = (proj.projections2[:, 2], proj.projections2[:, 2],
                     proj.projections2[:, 0], proj.projections2[:, 1],
                     2.0, 2.0, 1.0, 1.0,
                     2, 1, loc.locations2[:, 2])
            intensity2 = Intensity(cache2, proj)
            intensity2.compute()
            #
            snap2 = ImageGen(intensity2)
            snap2.snap(snap_num=2)
            # snap2.save_snap(fname=path + str(i) + '_2.tif')
            snap2.check_data(snap_num=2)

            print('Done with image 2 for pair number ' + str(i) + '\n')


if __name__ == '__main__':
    unittest.main()



import unittest
import matplotlib.pyplot as plt


class TestCCDProjection(unittest.TestCase):
    def test_ccd_projection(self):
        from src.sypivlib.function.dataio import GridIO, FlowIO
        from src.sypivlib.sypiv.create_particles import Particle, LaserSheet, CreateParticles
        from src.sypivlib.sypiv.ccd_projection import CCDProjection

        # Read-in the grid and flow file
        grid = GridIO('../data/plate_data/plate.sp.x')
        grid.read_grid()
        grid.compute_metrics()
        flow = FlowIO('../data/plate_data/sol-0000010.q')
        flow.read_flow()

        # Set particle data
        p = Particle()
        p.min_dia = 144e-9  # m
        p.max_dia = 573e-9  # m
        p.mean_dia = 281e-9  # m
        p.std_dia = 97e-9  # m
        p.density = 810  # kg/m3
        p.n_concentration = 50
        p.compute_distribution()
        # print(p.particle_field)

        # Read-in the laser sheet
        laser = LaserSheet(grid)
        laser.position = 0.05  # in m
        laser.thickness = 4e-3  # in m (Data obtained from LaVision)
        laser.pulse_time = 1e-5
        laser.compute_bounds()
        # print(laser.width)

        # Create particle locations array
        ia_bounds = [None, None, None, None]
        loc = CreateParticles(grid, flow, p, laser, ia_bounds)
        loc.ia_bounds = [0.3, 0.5, 0.3, 0.5]
        loc.in_plane = 70
        loc.compute_locations()
        loc.compute_locations2_serial()

        # Create particle projections (Simulating data from EUROPIV)
        proj = CCDProjection(loc)
        proj.d_ccd = 1000  # in mm
        proj.d_ia = 100  # in mm
        proj.compute()

        # Sample code to plot particle locations and relative diameters
        _in_plane = int(p.n_concentration * loc.in_plane * 0.01)
        # plot in-plane particle locations
        plt.scatter(proj.projections[:_in_plane, 0], proj.projections[:_in_plane, 1], c='g', s=5)
        # plot out-of-plane locations
        plt.scatter(proj.projections[_in_plane:, 0], proj.projections[_in_plane:, 1], c='r', s=5)

        plt.figure()
        # plot in-plane particle locations
        plt.scatter(proj.projections2[:_in_plane, 0], proj.projections2[:_in_plane, 1], c='g', s=5)
        # plot out-of-plane locations
        plt.scatter(proj.projections2[_in_plane:, 0], proj.projections2[_in_plane:, 1], c='r', s=5)

        plt.show()


if __name__ == '__main__':
    unittest.main()



# This file searches for the cell in which the given point is present in the grid

import numpy as np


# noinspection SpellCheckingInspection
class Search:
    """Module to search for the cell in which the given point is present

    ...

    Attributes
    ----------
    Input :
        grid : src.io.plot3dio.GridIO
            Grid object created from GridIO
        ppoint: list
            A float list of shape 3 -- Representing x, y, z of a point
    Output :
        cpoint: numpy.ndarray
            ppoint location in c-space. Will be computed if calculating in c-space
        index: numpy.ndarray
            Indices of closest node to the given point
        cell: numpy.ndarray
            Indices of the cell in which the given point is present
        info: str
            Information about the point location

    Methods
    -------
    compute()
        Finds the location of given point in the grid using the given search-method
        search-methods:
            distance, block-distance, p-space, c-space

    p2c()
        Method to convert location of point in physical space to computational space
        Rarely used

    c2p()
        Method to convert from c-space to p-space
        Used in integration algorithms to get new cell

    author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
    date: 10-24/2021
        """

    def __init__(self, grid, ppoint):
        self.grid = grid
        self.ppoint = ppoint
        self.cpoint = None
        self.index = None
        self.cell = None
        self.info = None
        self.block = None

    def __str__(self):
        doc = "This instance takes in the grid of shape " + self.grid.grd.shape + \
              "\nand the searches for the point " + self.ppoint + " in the grid.\n" \
              "Use method 'compute' to find (attributes) the closest 'index' and the nodes of the 'cell'.\n"
        return doc

    @staticmethod
    def _cell_nodes(_i, _j, _k):
        # _Internal method to get the nodes of a cell
        _cell = np.array([[_i, _j, _k],
                          [_i + 1, _j, _k],
                          [_i + 1, _j + 1, _k],
                          [_i, _j + 1, _k],
                          [_i, _j, _k + 1],
                          [_i + 1, _j, _k + 1],
                          [_i + 1, _j + 1, _k + 1],
                          [_i, _j + 1, _k + 1]], dtype=int)
        return _cell

    @staticmethod
    def _cell_index(self, i, j, k):
        # _Internal method to obtain the nodes of the cell in which the given point is present

        # Transform to found node to find the location of point
        # Basically looking at which quadrant the point is located
        # to find the nodes of the respective cell
        _node = self.grid.grd[i, j, k, :, self.block]
        _point_transform = self.ppoint - _node

        # Check if point is a node in the domain
        if np.all(abs(_point_transform) <= 1e-12):
            self.cell = self._cell_nodes(i, j, k)
            self.info = 'Given point is a node in the domain with a tol of 1e-12.\n' \
                        'Interpolation will assign node properties for integration.\n' \
                        'Index of the node will be returned by cell attribute\n'
            # print(self.info)
            return

        # ON BOUNDARY FOR A GENERALIZED HEXA IS SAME AS DEFAULT SEARCH
        # Removed the code for on the boundary case
        # Start the main cell nodes code
        if np.all(_point_transform >= 1e-6):
            self.cell = self._cell_nodes(i, j, k)
            return
        if _point_transform[0] <= 1e-6 and _point_transform[1] >= 1e-6 and _point_transform[2] >= 1e-6:
            self.cell = self._cell_nodes(i - 1, j, k)
            return
        if _point_transform[0] <= 1e-6 and _point_transform[1] <= 1e-6 and _point_transform[2] >= 1e-6:
            self.cell = self._cell_nodes(i - 1, j - 1, k)
            return
        if _point_transform[0] >= 1e-6 and _point_transform[1] <= 1e-6 and _point_transform[2] >= 1e-6:
            self.cell = self._cell_nodes(i, j - 1, k)
            return
        if _point_transform[0] >= 1e-6 and _point_transform[1] >= 1e-6 and _point_transform[2] <= 1e-6:
            self.cell = self._cell_nodes(i, j, k - 1)
            return
        if _point_transform[0] <= 1e-6 and _point_transform[1] >= 1e-6 and _point_transform[2] <= 1e-6:
            self.cell = self._cell_nodes(i - 1, j, k - 1)
            return
        if np.all(_point_transform <= 1e-6):
            self.cell = self._cell_nodes(i - 1, j - 1, k - 1)
            return
        if _point_transform[0] >= 1e-6 and _point_transform[1] <= 1e-6 and _point_transform[2] <= 1e-6:
            self.cell = self._cell_nodes(i, j - 1, k - 1)
            return

        return

    @staticmethod
    def _find_block(self):
        # _Internal method to find the block
        # Setup to compute block number in which the point is present
        _bool_min = self.grid.grd_min <= self.ppoint
        _bool_max = self.grid.grd_max >= self.ppoint
        _bool = _bool_max == _bool_min

        # Test if the given point is in domain or not
        if np.all(_bool.all(axis=1) == False) or np.all(_bool_min == False) or np.all(_bool_max == False):
            self.info = 'Given point is not in the domain. The cell attribute will return "None" in search algorithm\n'
            self.cell = None
            self.ppoint = None
            self.cpoint = None
            self.block = None
            # print(self.info)
            return
        # Assign the block number to the attribute
        self.block = int(np.where(_bool.all(axis=1))[0][0])

        return self.block

    def compute(self, method='block_distance'):
        """
        Use the method to compute index and cell attributes

        parameter:
            method: str
                Currently distance or block_distance

        return:
        None

        author: Dilip Kalagotla @ kal ~ dilip.kalagotla@gmail.com
        date: 10-24/2021
        """

        # Find the block number
        self.block = self._find_block(self)
        # To check for point out-of-domain case
        if self.block is None:
            return

        match method:

            case 'distance':
                # Compute the distance from all nodes in the grid
                _dist = np.sqrt((self.grid.grd[..., 0, :] - self.ppoint[0]) ** 2 +
                                (self.grid.grd[..., 1, :] - self.ppoint[1]) ** 2 +
                                (self.grid.grd[..., 2, :] - self.ppoint[2]) ** 2)

                # Find the closest node to the point --> index.ndim = 4
                self.index = np.array(np.unravel_index(_dist.argmin(), _dist.shape))
                i, j, k, self.block = self.index[0], self.index[1], self.index[2], self.index[3]
                self._cell_index(self, i, j, k)
                # Check for the end of the domain case
                if max(self.cell[:, 0]) > self.grid.ni[self.block] - 1 or \
                        max(self.cell[:, 1]) > self.grid.nj[self.block] - 1 or \
                        max(self.cell[:, 2]) > self.grid.nk[self.block] - 1:
                    self.cpoint = None
                    self.ppoint = None
                    return

            case 'block_distance':
                # Compute distance inside the block to get the nearest node
                _i, _j, _k = self.grid.ni[self.block], self.grid.nj[self.block], self.grid.nk[self.block]
                _dist = np.sqrt((self.grid.grd[:_i, :_j, :_k, 0, self.block] - self.ppoint[0]) ** 2 +
                                (self.grid.grd[:_i, :_j, :_k, 1, self.block] - self.ppoint[1]) ** 2 +
                                (self.grid.grd[:_i, :_j, :_k, 2, self.block] - self.ppoint[2]) ** 2)

                # Other methods to calculate distance. The above method is faster
                # _grd_min_point = self.grid.grd[:_i, :_j, :_k, :, self.block] - self.point
                # _dist = np.linalg.norm(_grd_min_point, axis=-1)
                # _dist = np.sqrt(np.einsum("ijkl,ijkl->ijk", _grd_min_point, _grd_min_point))

                self.index = np.array(np.unravel_index(_dist.argmin(), _dist.shape))
                i, j, k = self.index
                self._cell_index(self, i, j, k)
                # Check for the end of the domain case
                if max(self.cell[:, 0]) > self.grid.ni[self.block] - 1 or \
                        max(self.cell[:, 1]) > self.grid.nj[self.block] - 1 or \
                        max(self.cell[:, 2]) > self.grid.nk[self.block] - 1 or np.any(self.cell < 0):
                    print('Block search returned wrong cell! Point position lost.\n')
                    self.cpoint = None
                    self.ppoint = None
                    return

            case 'p-space':
                # Search for given point using newton-raphson
                # credit: Sadarjoen et al.
                # title: Particle tracing algorithms for 3D Curvilinear grids
                # This search is performed every single time to find the given point
                self.cpoint = self.p2c(self.ppoint)

            case 'c-space':
                # To run c-space global point location is needed
                # credit: Sadarjoen et al.
                # title: Particle tracing algorithms for 3D Curvilinear grids

                # Transform given point from p-space to c-space using newton-raphson
                # This is only performed once to get the initial c-space point
                self.cpoint = self.p2c(self.ppoint)

    def c2p(self, _cpoint):
        """
        Method to convert c-space point to p-space
        self.block is kept constant through the c-space algos
        self.block is checked and switched in streamlines algo
        This method is commonly used to test point location as well
        in different algorithms

        Args:
            _cpoint: c-space co-ordinates

        Returns:
            _ppoint: p-space co-ordinates

        """
        self.cpoint = _cpoint
        _eps0, _eps1, _eps2 = _cpoint.astype(int)
        _alpha, _beta, _gamma = np.modf(_cpoint)[0]  # same as eps % 1.0

        # Check if the given point is in the domain
        if _eps0 >= self.grid.ni[self.block]-1 or _eps1 >= self.grid.nj[self.block]-1 or \
                _eps2 >= self.grid.nk[self.block]-1:
            self.cpoint = None
            self.ppoint = None
            return self.ppoint

        # Determine in which cell the current point is present
        self.cell = self._cell_nodes(_eps0, _eps1, _eps2)
        _cell_grd = self.grid.grd[self.cell[:, 0], self.cell[:, 1], self.cell[:, 2], :, self.block]

        # Calculate the location in p-space
        self.ppoint = (1 - _alpha) * (1 - _beta) * (1 - _gamma) * _cell_grd[0] + \
                _alpha  * (1 - _beta) * (1 - _gamma) * _cell_grd[1] + \
                _alpha  *      _beta  * (1 - _gamma) * _cell_grd[2] + \
                (1 - _alpha) *      _beta  * (1 - _gamma) * _cell_grd[3] + \
                (1 - _alpha) * (1 - _beta) *      _gamma  * _cell_grd[4] + \
                _alpha  * (1 - _beta) *      _gamma  * _cell_grd[5] + \
                _alpha  *      _beta *       _gamma  * _cell_grd[6] + \
                (1 - _alpha) *      _beta  *      _gamma  * _cell_grd[7]

        return self.ppoint

    def p2c(self, _ppoint):
        """
        Method to convert p-space point to c-space
        As there is no direct analytical equation. We use Newton-Raphson
        Args:
            _ppoint: p-space co-ordinates

        Returns:
            eps: c-space co-ordinates
        """
        global _cpoint
        self.ppoint = _ppoint

        if self.block is None:
            self.block = self._find_block(self)

        # Start Newton-Raphson
        _iter = 0
        # Initial guess
        try:
            _cpoint is not None
        except:
            _cpoint = np.array([self.grid.ni[self.block], self.grid.nj[self.block], self.grid.nk[self.block]]) / 2 + \
                np.random.randn(3)

        while True:
            # Check if taking too long
            if _iter >= 1e3:
                print('**ERROR** Newton-Raphson did not converge. Try again!\n'
                      'Possible reason might be the point might be too close to the end of a domain')
                self.ppoint, self.cpoint = None, None
                return

            # Check for out-of-domain case and reset the point to in-domain
            _eps0, _eps1, _eps2 = _cpoint.astype(int)
            _alpha, _beta, _gamma = np.modf(_cpoint)[0]
            if _eps0 + 1 >= self.grid.ni[self.block]:
                _cpoint[0] = self.grid.ni[self.block] - 1 - _alpha
            if _eps1 + 1 >= self.grid.nj[self.block]:
                _cpoint[1] = self.grid.nj[self.block] - 1 - _beta
            if _eps2 + 1 >= self.grid.nk[self.block]:
                _cpoint[2] = self.grid.nk[self.block] - 1 - _gamma

            # Compute eps after the point is reset
            _eps0, _eps1, _eps2 = _cpoint.astype(int)
            self.cell = self._cell_nodes(_eps0, _eps1, _eps2)

            # Calculate J_inv for the point --> Uses tri-linear interpolation
            _cell_J_inv = self.grid.m2[self.cell[:, 0], self.cell[:, 1], self.cell[:, 2], :, :, self.block]
            _J_inv = (1 - _alpha) * (1 - _beta) * (1 - _gamma) * _cell_J_inv[0] + \
                     _alpha * (1 - _beta) * (1 - _gamma) * _cell_J_inv[1] + \
                     _alpha * _beta * (1 - _gamma) * _cell_J_inv[2] + \
                     (1 - _alpha) * _beta * (1 - _gamma) * _cell_J_inv[3] + \
                     (1 - _alpha) * (1 - _beta) * _gamma * _cell_J_inv[4] + \
                     _alpha * (1 - _beta) * _gamma * _cell_J_inv[5] + \
                     _alpha * _beta * _gamma * _cell_J_inv[6] + \
                     (1 - _alpha) * _beta * _gamma * _cell_J_inv[7]

            # Transform from c to p-space
            _pred_ppoint = self.c2p(_cpoint)

            # Difference b/w predicted point to given point
            _delta_ppoint = _ppoint - _pred_ppoint

            # End newton-raphson if condition is met
            # TODO: Condition needs to be adapted based on Jacobian
            # TODO: Need to improve by normalizing the data
            _tol = 1e-12 * self.grid.J[self.cell[0, 0], self.cell[0, 1], self.cell[0, 2], self.block]
            if _tol <= 1e-12:
                _tol = 1e-12
            if sum(abs(_delta_ppoint)) <= _tol:
                _eps0, _eps1, _eps2 = _cpoint.astype(int)
                # same as self.cell = self._cell_nodes(_eps0, _eps1, _eps2)
                self._cell_index(self, _eps0, _eps1, _eps2)
                self.cpoint = _cpoint
                self.ppoint = _pred_ppoint
                return _cpoint

            # Transform from p to c-space
            _delta_cpoint = np.matmul(_J_inv, _delta_ppoint)

            # Save old point
            _cpoint_old = _cpoint.copy()

            # Update point
            _cpoint += _delta_cpoint
            # Update the point to zero if less than zero
            _cpoint[_cpoint < 0] = 0
            _cpoint = abs(_cpoint)
            _iter += 1

            pass





import numpy as np


class CCDProjection:
    """
    Use to project particles obtained from create_particles onto
    CCD of a camera system
    """

    def __init__(self, particles, xres=512, yres=512, dpi=96):
        self.particles = particles
        self.d_ccd = None
        self.d_ia = None
        self.xres = xres
        self.yres = yres
        self.dpi = dpi
        self.projections = np.empty((self.particles.locations.shape[0], 3))
        self.projections2 = np.empty((self.particles.locations2.shape[0], 3))

    def compute(self):
        """
        Computes the particles projection onto a CCD
        Set the sensor size using xres, yres, and dpi
        Appropriate distances for projection should be inputted as well (d_ccd & d_ia)
        :return:
        projections: np.ndarray of shape (nx3)
        projections2: np.ndarray of shape (nx3)
        (x,y,d) --> locations of particles in pixels and their respective diameters in meters
        """
        loc = self.particles.locations
        origin = [np.mean(self.particles.ia_bounds[:2]), np.mean(self.particles.ia_bounds[2:]), loc[0, 2]]
        # Adjust locations to origin before projection
        loc_adj = loc[:, :3] - origin
        self.projections[:, 0] = loc_adj[:, 0] * self.d_ccd / (loc_adj[:, 2] - self.d_ia)
        self.projections[:, 1] = loc_adj[:, 1] * self.d_ccd / (loc_adj[:, 2] - self.d_ia)
        self.projections[:, 2] = loc[:, 3]

        # Convert to pixels for further processing
        self.projections[:, :2] = self.projections[:, :2] * self.dpi / 25.4e-3

        # Compute projections for the second snap
        loc = self.particles.locations2
        origin = [np.mean(self.particles.ia_bounds[:2]), np.mean(self.particles.ia_bounds[2:]), loc[0, 2]]
        # Adjust locations to origin before projection
        loc_adj = loc[:, :3] - origin
        self.projections2[:, 0] = loc_adj[:, 0] * self.d_ccd / (loc_adj[:, 2] - self.d_ia)
        self.projections2[:, 1] = loc_adj[:, 1] * self.d_ccd / (loc_adj[:, 2] - self.d_ia)
        self.projections2[:, 2] = loc[:, 3]

        # Convert to pixels for further processing
        self.projections2[:, :2] = self.projections2[:, :2] * self.dpi / 25.4e-3

        return

    pass



import unittest
import matplotlib.pyplot as plt
import numpy as np


class TestIntensity(unittest.TestCase):
    def test_intensity(self):
        from src.sypivlib.function.dataio import GridIO, FlowIO
        from src.sypivlib.sypiv.create_particles import Particle, LaserSheet, CreateParticles
        from src.sypivlib.sypiv.ccd_projection import CCDProjection
        from src.sypivlib.sypiv.intensity import Intensity

        # Read-in the grid and flow file
        grid = GridIO('../data/plate_data/plate.sp.x')
        grid.read_grid()
        grid.compute_metrics()
        flow = FlowIO('../data/plate_data/sol-0000010.q')
        flow.read_flow()

        # Set particle data
        p = Particle()
        p.min_dia = 144e-9  # m
        p.max_dia = 573e-9  # m
        p.mean_dia = 281e-9  # m
        p.std_dia = 97e-9  # m
        p.density = 810  # kg/m3
        p.n_concentration = 25
        p.compute_distribution()

        # Read-in the laser sheet
        laser = LaserSheet(grid)
        laser.position = 0.05  # in m
        laser.thickness = 4e-3  # in m (Data obtained from LaVision)
        laser.pulse_time = 4e-6
        laser.compute_bounds()

        # Create particle locations array
        ia_bounds = [None, None, None, None]
        loc = CreateParticles(grid, flow, p, laser, ia_bounds)
        loc.ia_bounds = [0.3, 0.5, 0.3, 0.5]  # in m
        loc.in_plane = 70
        loc.compute_locations()
        loc.compute_locations2_serial()

        # Sample code to plot particle locations and relative diameters
        _in_plane = int(p.n_concentration * loc.in_plane * 0.01)
        # plot in-plane particle locations
        plt.scatter(loc.locations[:_in_plane, 0], loc.locations[:_in_plane, 1],
                    s=10 * loc.locations[:_in_plane, 3] / p.min_dia, c='g')
        # plot out-of-plane locations
        plt.scatter(loc.locations[_in_plane:, 0], loc.locations[_in_plane:, 1],
                    s=10 * loc.locations[_in_plane:, 3] / p.min_dia, c='r')

        # Create particle projections (Simulating data from EUROPIV)
        proj = CCDProjection(loc)
        proj.d_ccd = 900  # in m
        proj.d_ia = 1000  # in m
        proj.dpi = 72
        proj.compute()

        # Sample code to plot particle locations and relative diameters
        _in_plane = int(p.n_concentration * loc.in_plane * 0.01)
        # plot in-plane particle locations
        plt.figure()
        plt.scatter(proj.projections[:_in_plane, 0], proj.projections[:_in_plane, 1], c='g',
                    s=10 * proj.projections[:_in_plane, 2] / p.min_dia)
        # plot out-of-plane locations
        plt.scatter(proj.projections[_in_plane:, 0], proj.projections[_in_plane:, 1], c='r',
                    s=10 * proj.projections[_in_plane:, 2] / p.min_dia)
        plt.title('Projected data - Not scaled')
        cache = (proj.projections[:, 2], proj.projections[:, 2],
                 proj.projections[:, 0], proj.projections[:, 1],
                 0.5, 0.5, 1.0, 1.0,
                 2, 1, loc.locations[:, 2])
        intensity = Intensity(cache, proj)
        intensity.compute_serial()

        # Creating temp arrays to test. This will be done internally in the code of image_gen
        xp, yp = proj.projections[:, 0], proj.projections[:, 1]
        x = np.linspace(-proj.xres / 2, proj.xres / 2, proj.xres)
        y = np.linspace(-proj.yres / 2, proj.yres / 2, proj.yres)
        xsize = proj.xres / proj.dpi
        ysize = proj.yres / proj.dpi

        # Plot the contour
        fig = plt.figure(figsize=[xsize, ysize], dpi=proj.dpi)
        ax = plt.axes([0.0, 0.0, 1.0, 1.0], xlim=(x.min(), x.max()), ylim=(y.min(), y.max()))
        ax.contourf(x, y, intensity.values, cmap='hot')

        # Plot another to look at the scale
        plt.figure()
        plt.contourf(x, y, intensity.values)

        # plot image to finalize the image
        fig = plt.figure(figsize=[xsize, ysize], dpi=proj.dpi)
        ax = plt.axes([0.0, 0.0, 1.0, 1.0])
        ax.imshow(intensity.values, cmap='gray', origin='lower')
        plt.show()


if __name__ == '__main__':
    unittest.main()
